# file: widget.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the widget class.

from __future__ import absolute_import

import pygame
from numpy import array
from inspect import getmembers

from singularity.code import g
from singularity.code.graphics import g as gg, constants

# surface.blits is available in 1.9.4 but it is not really useful until
# 1.9.5 or 1.9.6.  We pick 1.9.6 to be sure it works.
if pygame.version.vernum > (1, 9, 5):
    HAS_FUNCTIONAL_BLITS = True
else:
    HAS_FUNCTIONAL_BLITS = False


def unmask(widget):
    """Causes the widget to exist above its parent's fade mask.  The widget's
    children will still be masked, unless they are unmasked themselves."""
    unmask_all(widget)
    widget.mask_children = True


def unmask_all(widget):
    """Causes the widget to exist above its parent's fade mask.  The widget's
    children will not be masked."""
    widget.self_mask = True
    widget.do_mask = lambda: None


def call_on_change(data_member, call_me, *args, **kwargs):
    """Creates a data member that calls a function when changed."""

    def get(self):
        return getattr(self, data_member)

    def set(self, my_value):
        if data_member in self.__dict__:
            change = my_value != self.__dict__[data_member]
        else:
            change = True

        if change:
            setattr(self, data_member, my_value)
            call_me(self, *args, **kwargs)

    return property(get, set)


def set_on_change(data_member, set_me, set_value=True):
    """Creates a data member that sets another data member to a given value
    when changed."""
    return call_on_change(data_member, setattr, set_me, set_value)


def causes_rebuild(data_member):
    """Creates a data member that sets needs_rebuild to True when changed."""
    return set_on_change(data_member, "needs_rebuild")


def causes_redraw(data_member):
    """Creates a data member that sets needs_redraw to True when changed."""
    return set_on_change(data_member, "needs_redraw")


def causes_resize(data_member):
    """Creates a data member that sets needs_resize to True when changed."""
    return set_on_change(data_member, "needs_resize")


def causes_reposition(data_member):
    """Creates a data member that sets needs_reposition to True when changed."""
    return set_on_change(data_member, "needs_reposition")


def causes_update(data_member):
    """Creates a data member that sets needs_update to True when changed."""
    return set_on_change(data_member, "needs_update")


def propogate_need(data_member):
    """Creates a function that can be passed to call_on_change.  When the
    data member changes to True, needs_update is set, and the True value
    is passed to all descendants."""

    def do_propogate(self):
        if getattr(self, data_member, False):
            self.needs_update = True

            if hasattr(self, "children"):
                descendants = self.children[:]
                while descendants:
                    child = descendants.pop()
                    # Propogate to this child and its descendants, if needed.
                    if not getattr(child, data_member, False):
                        setattr(child, data_member, True)
                        child._needs_update = True
                        if hasattr(child, "children"):
                            descendants += child.children

    return do_propogate


# Previous attempt was to hide the raw value by resolving
# the value before returning it. However, there are legitimate
# reason to access the raw value. So, we need two property.
# Using a wrapper is not worth the trouble.
# I choose to let the unresolved value the default because
# in majority of it's what we want, and in other case,
# you must handle reconfig anyways.
class auto_reconfig(object):
    __slots__ = [
        "data_member",
        "reconfig_datamember",
        "reconfig_func",
    ]  # Avoid __dict__.

    def __init__(self, data_member, reconfig_prefix, reconfig_func):
        self.data_member = data_member
        self.reconfig_datamember = reconfig_prefix + data_member
        self.reconfig_func = reconfig_func

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return getattr(obj, self.data_member)

    def __set__(self, obj, my_value):
        new_value = self.reconfig_func(my_value)

        setattr(obj, self.reconfig_datamember, new_value)
        setattr(obj, self.data_member, my_value)

    def reconfig(self, obj):
        updated_value = self.reconfig_func(getattr(obj, self.data_member))
        setattr(obj, self.reconfig_datamember, updated_value)


# In debug mode, this list tracks which widgets (i.e. rects) where highlighted "last"
# during a redraw, so they can be "re-updated" without the highlight
debug_mode_undo_drawing_highlight = []


class Widget(object):
    """A Widget is a GUI element.  It can have one parent and any number of
    children."""

    needs_redraw = call_on_change("_needs_redraw", propogate_need("_needs_redraw"))

    needs_resize = call_on_change("_needs_resize", propogate_need("_needs_resize"))

    needs_reposition = call_on_change(
        "_needs_reposition", propogate_need("_needs_reposition")
    )

    needs_rebuild = causes_update("_needs_rebuild")

    def _propogate_update(self):
        if self._needs_update:
            if hasattr(self, "parent"):
                target = self.parent
                while target and not target._needs_update:
                    target._needs_update = True
                    target = target.parent

    needs_update = call_on_change("_needs_update", _propogate_update)

    needs_reconfig = call_on_change(
        "_needs_reconfig", propogate_need("_needs_reconfig")
    )

    pos = causes_reposition("_pos")
    size = causes_resize("_size")
    anchor = causes_reposition("_anchor")
    visible = causes_redraw("_visible")

    def __init__(self, parent, pos, size, anchor=constants.TOP_LEFT):
        self.parent = parent
        self.children = []

        self.pos = pos
        self.size = size
        self.anchor = anchor

        # "It's a widget!"
        self.add_hooks()

        self.is_above_mask = False
        self.self_mask = False
        self.mask_children = False
        self.visible = True

        self.needs_rebuild = True
        self.collision_rect = None

        # Set automatically by other properties.
        # self.needs_redraw = True
        # self.needs_full_redraw = True
        self.needs_reconfig = True

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        if hasattr(self, "children"):
            self.remove_hooks()

        if hasattr(self, "_parent") and self._parent is not None:
            try:
                self._parent.children.remove(self)
            except ValueError:
                pass  # Wasn't there to start with.

        self._parent = parent

        if self.parent is not None:
            self.parent.children.append(self)
            self.parent.needs_rebuild = True
            self.parent.needs_resize = True
            self.parent.needs_reposition = True
            self.parent.needs_redraw = True

        if hasattr(self, "children"):
            self.add_hooks()

    def add_hooks(self):
        if self.parent is not None:
            # Won't trigger on the call from __init__, since there are no
            # children yet, but add_hooks may be explicitly called elsewhere to
            # undo remove_hooks.
            for child in self.children:
                child.add_hooks()

    def remove_hooks(self):
        # Localize the children list to avoid index corruption and O(N^2) time.
        children = self.children

        # Recurse to the children.
        for child in children:
            child.remove_hooks()

    def _parent_size(self):
        if self.parent == None:
            return gg.real_screen_size
        else:
            return self.parent.real_size

    def _calc_size(self):
        """Internal method.  Calculates and returns the real size of this
        widget.

        Override to create a dynamically-sized widget."""
        parent_size = self._parent_size()
        size = list(self.size)
        for i in range(2):
            if size[i] > 0:
                size[i] = int(size[i] * gg.real_screen_size[i])
            elif size[i] < 0:
                size[i] = int((-size[i]) * parent_size[i])

        return tuple(size)

    def get_real_size(self):
        """Returns the real size of this widget.

        To implement a dynamically-sized widget, override _calc_size, which
        will be called whenever the widget is resized, and set needs_resize
        when appropriate."""
        return self._real_size

    real_size = property(get_real_size)

    def get_real_pos(self):
        """Returns the real position of this widget on its parent."""
        vanchor, hanchor = self.anchor
        parent_size = self._parent_size()
        my_size = self.real_size

        if self.pos[0] >= 0:
            hpos = int(self.pos[0] * gg.real_screen_size[0])
        else:
            hpos = -int(self.pos[0] * parent_size[0])

        if hanchor == constants.LEFT:
            pass
        elif hanchor == constants.CENTER:
            hpos -= my_size[0] // 2
        elif hanchor == constants.RIGHT:
            hpos -= my_size[0]

        if self.pos[1] >= 0:
            vpos = int(self.pos[1] * gg.real_screen_size[1])
        else:
            vpos = -int(self.pos[1] * parent_size[1])

        if vanchor == constants.TOP:
            pass
        elif vanchor == constants.MID:
            vpos -= my_size[1] // 2
        elif vanchor == constants.BOTTOM:
            vpos -= my_size[1]

        return (hpos, vpos)

    real_pos = property(get_real_pos)

    def _make_collision_rect(self):
        """Creates and returns a collision rect for this widget."""
        pos = array(self.real_pos)
        if self.parent:
            pos += self.parent.collision_rect[:2]

        return pygame.Rect(pos, self.real_size)

    def is_over(self, position):
        if not getattr(self, "collision_rect", None):
            return False

        if position != (0, 0):
            return self.collision_rect.collidepoint(position)
        else:
            return False

    def remake_surfaces(self):
        """Recreates the surfaces that this widget will draw on."""
        size = self.real_size
        pos = self.real_pos

        if self.parent != None:
            try:
                self.surface = self.parent.surface.subsurface(pos + size)
            except ValueError:
                print("Warning: %r can't fit on its parent." % self)
                print(pos, size, self.parent.real_pos, self.parent.real_size)

                wanted_rect = pos + size
                available_rect = self.parent.surface.get_rect()
                compromise = available_rect.clip(wanted_rect)

                self.surface = self.parent.surface.subsurface(compromise)
        else:
            # Recreate using the abstracted screen size, NOT the real one
            # g.set_screen() will calculate the proper g.real_screen_size
            if gg.screen_surface is None:
                # Ensure that the screen is initialized
                gg.set_mode()
            # We draw on a copy of the surface.  This is to avoid crashes
            # during draggable resizing (event.VIDEORESIZE) where the
            # screen size might change behind our backs while drawing
            # (event.VIDEORESIZE tells us that the screen has been updated
            # and we should catch up and not the other way around)
            self.surface = gg.screen_surface.copy()
            self.surface.fill((0, 0, 0, 255))

            gg.fade_mask = pygame.Surface(size, 0, gg.ALPHA)
            gg.fade_mask.fill((0, 0, 0, 175))

    def prepare_for_redraw(self):
        # First, we handle config changes.
        if self.needs_reconfig:
            self.reconfig()
            self.needs_reconfig = False

        # Then any substance changes.
        if self.needs_rebuild:
            self.rebuild()
            self.needs_rebuild = False

        # Then size changes.
        if self.needs_resize:
            self.resize()
            self.needs_resize = False
            self.needs_reposition = True
            self.needs_redraw = True

        # Then position changes.
        if self.needs_reposition:
            self.needs_reposition = False
            self.reposition()

        # And finally we recurse to our descendants.
        for child in self.children:
            if child.visible:
                child.prepare_for_redraw()

    def maybe_update(self):
        if self.needs_update:
            self.update()

    def update(self):
        # First we prepare everything for its redraw (if needed).
        self.prepare_for_redraw()

        _, updated_rect = self._update()

        # Oh, and if this is the top-level widget, we should update the display.
        if not self.parent and gg.screen_surface:
            root_surface = self.surface
            if g.debug and updated_rect:
                # In debug mode, draw red boxes to represent widgets that were updated.
                global debug_mode_undo_drawing_highlight
                try:
                    # If the theme defines a color for this purpose, we will use it
                    widget_highlight_color = gg.resolve_color_alias(
                        "debug_mode_highlight_redrawn_widget"
                    )
                except KeyError:
                    # ... and for every thing else, there is the color red.
                    widget_highlight_color = 0xFF, 0, 0, 0
                root_surface = self.surface.copy()
                n_updated_rect = []
                for rect in updated_rect:
                    n_updated_rect.append(
                        pygame.draw.rect(root_surface, widget_highlight_color, rect, 1)
                    )
                updated_rect.extend(debug_mode_undo_drawing_highlight)
                debug_mode_undo_drawing_highlight = n_updated_rect

            if HAS_FUNCTIONAL_BLITS:
                gg.screen_surface.blits(
                    ((root_surface, r, r) for r in updated_rect), doreturn=0
                )
            else:
                for r in updated_rect:
                    gg.screen_surface.blit(root_surface, r, area=r)

            pygame.display.update(updated_rect)

    def _update(self):
        redrew_self = self.needs_redraw
        update_full_rect = redrew_self
        affected_rects = []
        if self.needs_redraw:
            self.redraw()

        # Then we update any children below our fade mask.
        check_mask = []
        above_mask = []
        for child in self.children:
            if child.needs_update and child.visible:
                if child.is_above_mask:
                    above_mask.append(child)
                else:
                    # update_full_rect = True  # We do not bother tracking this case
                    child_mask, child_rects = child._update()
                    check_mask.extend(child_mask)
                    if child_rects and not update_full_rect:
                        affected_rects.extend(child_rects)

        # Next, we handle the fade mask.
        if getattr(self, "faded", False):
            while check_mask:
                child = check_mask.pop()
                if not child.self_mask:
                    # update_full_rect = True  # We do not bother tracking this case
                    child_rect = child.surface.blit(gg.fade_mask, (0, 0))
                    if not update_full_rect:
                        affected_rects.append(child_rect)
                elif child.mask_children:
                    check_mask += child.children

        # And finally we update any children above the fade mask.
        for child in above_mask:
            _, child_rects = child._update()
            if child_rects and not update_full_rect:
                affected_rects.extend(child_rects)

        # Update complete.
        self.needs_update = False

        # Any descendants we didn't check for masking get passed upwards.
        if redrew_self:
            # If we redrew this widget, we tell our parent to consider it
            # instead.  The parent will recurse down to any descendants if
            # needed, and redraw already propagated down to them.
            check_mask = [self]

        if update_full_rect:
            affected_rects = [self.collision_rect]

        return check_mask, affected_rects

    def reconfig(self):
        # Find reconfig property and update them.
        clazz = self.__class__
        for prop_name, prop in getmembers(clazz):
            if isinstance(prop, auto_reconfig):
                prop.reconfig(self)

    def rebuild(self):
        pass

    def resize(self):
        self._real_size = self._calc_size()

    def reposition(self):
        old_rect = self.collision_rect
        self.collision_rect = self._make_collision_rect()

        if not self.parent:
            self.remake_surfaces()
            self.needs_redraw = True
        elif (
            (getattr(self, "surface", None) is None)
            or (old_rect is None)
            or (self.surface.get_parent() is not self.parent.surface)
            or (not self.collision_rect.contains(old_rect))
        ):
            self.remake_surfaces()
            self.parent.needs_redraw = True
        elif self.collision_rect != old_rect:
            self.remake_surfaces()
            self.needs_redraw = True

    def redraw(self):
        self.needs_redraw = False
        if self.parent is None:
            self.surface.fill((0, 0, 0, 255))

    def add_handler(self, *args, **kwargs):
        """Handler pass-through."""
        if self.parent:
            self.parent.add_handler(*args, **kwargs)

    def remove_handler(self, *args, **kwargs):
        """Handler pass-through."""
        if self.parent:
            self.parent.remove_handler(*args, **kwargs)

    def add_key_handler(self, *args, **kwargs):
        """Handler pass-through."""
        if self.parent:
            self.parent.add_key_handler(*args, **kwargs)

    def remove_key_handler(self, *args, **kwargs):
        """Handler pass-through."""
        if self.parent:
            self.parent.remove_key_handler(*args, **kwargs)

    def add_focus_widget(self, *args, **kwargs):
        """Focus pass-through."""
        if self.parent:
            self.parent.add_focus_widget(*args, **kwargs)

    def remove_focus_widget(self, *args, **kwargs):
        """Focus pass-through."""
        if self.parent:
            self.parent.remove_focus_widget(*args, **kwargs)

    def took_focus(self, *args, **kwargs):
        """Focus pass-through."""
        if self.parent:
            self.parent.took_focus(*args, **kwargs)

    def clear_focus(self, *args, **kwargs):
        """Focus pass-through."""
        if self.parent:
            self.parent.clear_focus(*args, **kwargs)


class BorderedWidget(Widget):
    borders = causes_redraw("__borders")

    border_color = auto_reconfig("_border_color", "resolved", gg.resolve_color_alias)
    background_color = auto_reconfig(
        "_background_color", "resolved", gg.resolve_color_alias
    )
    resolved_border_color = causes_redraw("_resolved_border_color")
    resolved_background_color = causes_redraw("_resolved_background_color")

    def __init__(self, parent, *args, **kwargs):
        self.borders = kwargs.pop("borders", ())
        self.border_color = kwargs.pop("border_color", "widget_border")
        self.background_color = kwargs.pop("background_color", "widget_background")

        super(BorderedWidget, self).__init__(parent, *args, **kwargs)

    def rebuild(self):
        super(BorderedWidget, self).rebuild()
        if self.parent and self.resolved_background_color == gg.colors["clear"]:
            self.parent.needs_redraw = True

    def reposition(self):
        super(BorderedWidget, self).reposition()
        if self.parent and self.resolved_background_color == gg.colors["clear"]:
            self.parent.needs_redraw = True

    def redraw(self):
        super(BorderedWidget, self).redraw()

        # TODO: Transparency do not work correctly.
        # First: fill cannot use alpha channel with current surface.
        # Second: Transparency needs the parent redraw to work correctly.
        # It make transparency unusable with some widget.

        # Fill the background.
        if self.resolved_background_color != gg.colors["clear"]:
            self.surface.fill(self.resolved_background_color)

        self.draw_borders()

    def draw_borders(self):
        my_size = self.real_size

        for edge in self.borders:
            if edge == constants.TOP:
                self.surface.fill(self.resolved_border_color, (0, 0, my_size[0], 1))
            elif edge == constants.LEFT:
                self.surface.fill(self.resolved_border_color, (0, 0, 1, my_size[1]))
            elif edge == constants.RIGHT:
                self.surface.fill(
                    self.resolved_border_color, (my_size[0] - 1, 0) + my_size
                )
            elif edge == constants.BOTTOM:
                self.surface.fill(
                    self.resolved_border_color, (0, my_size[1] - 1) + my_size
                )


class FocusWidget(Widget):
    has_focus = causes_redraw("_has_focus")

    def __init__(self, *args, **kwargs):
        super(FocusWidget, self).__init__(*args, **kwargs)
        self.has_focus = False

        self.add_handler(constants.CLICK, self.handle_click, 0)

    def add_hooks(self):
        super(FocusWidget, self).add_hooks()
        if self.parent is not None:
            self.parent.add_focus_widget(self)

    def remove_hooks(self):
        super(FocusWidget, self).remove_hooks()
        if self.parent is not None:
            self.parent.remove_focus_widget(self)

    def handle_click(self, event):
        if not self.is_over(event.pos):
            self.clear_focus(self)



# file: slider.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the slider widget.

from __future__ import absolute_import

import pygame

from singularity.code.graphics import g, constants, widget, button


def calc_max(elements, size):
    return max(elements - size, 0)


class Slider(button.Button):
    slider_pos = widget.causes_rebuild("_slider_pos")
    slider_max = widget.causes_rebuild("_slider_max")
    slider_size = widget.causes_rebuild("_slider_size")
    horizontal = widget.causes_rebuild("_horizontal")

    slider_color = widget.auto_reconfig(
        "_slider_color", "resolved", g.resolve_color_alias
    )
    resolved_slider_color = widget.causes_redraw("_resolved_slider_color")

    def __init__(
        self,
        parent,
        pos=(-1, 0),
        size=(-0.1, -1),
        anchor=constants.TOP_RIGHT,
        borders=constants.ALL,
        border_color=None,
        background_color=None,
        slider_color=None,
        slider_pos=0,
        slider_max=10,
        slider_size=5,
        horizontal=False,
        **kwargs
    ):
        kwargs.setdefault("priority", 80)
        super(Slider, self).__init__(parent, pos, size, anchor=anchor, **kwargs)

        border_color = border_color or "slider_border"
        background_color = background_color or "slider_background"
        slider_color = slider_color or "slider_background_slider"

        self.borders = borders
        self.border_color = border_color
        self.background_color = background_color
        self.selected_color = background_color
        self.unselected_color = background_color
        self.slider_color = slider_color

        self.slider_pos = slider_pos
        self.slider_max = slider_max
        self.slider_size = slider_size
        self.horizontal = horizontal

        self.drag_state = None
        self.button = button.Button(
            self,
            pos=None,
            size=None,
            anchor=constants.TOP_LEFT,
            border_color=border_color,
            selected_color=slider_color,
            unselected_color=slider_color,
            priority=self.priority - 5,
        )

    def redraw(self):
        super(Slider, self).redraw()
        self.button.selected_color = self.slider_color
        self.button.unselected_color = self.slider_color

    def add_hooks(self):
        super(Slider, self).add_hooks()
        self.parent.add_handler(constants.DRAG, self.handle_drag)
        self.parent.add_handler(constants.CLICK, self.handle_click, 50)

    def remove_hooks(self):
        super(Slider, self).remove_hooks()
        self.parent.remove_handler(constants.DRAG, self.handle_drag)
        self.parent.remove_handler(constants.CLICK, self.handle_click)

    def _calc_length(self, items):
        return items / float(self.slider_size + self.slider_max)

    def rebuild(self):
        super(Slider, self).rebuild()
        self.needs_resize = True

    def resize(self):
        super(Slider, self).resize()
        bar_start = self._calc_length(self.slider_pos)
        bar_length = self._calc_length(self.slider_size)

        if self.horizontal:
            self.button.pos = (-bar_start, 0)
            self.button.size = (-bar_length, -1)
            borders = [constants.TOP, constants.BOTTOM]

            self.button.resize()
            real_pos = self.button.real_pos[0]
            real_size = self.button.real_size[0]
            if real_pos == 0:
                borders.append(constants.LEFT)
            if real_pos + real_size == self.real_size[0]:
                borders.append(constants.RIGHT)
            self.button.borders = tuple(borders)
        else:
            self.button.pos = (0, -bar_start)
            self.button.size = (-1, -bar_length)
            borders = [constants.LEFT, constants.RIGHT]

            self.button.resize()
            real_pos = self.button.real_pos[1]
            real_size = self.button.real_size[1]
            if real_pos == 0:
                borders.append(constants.TOP)
            if real_pos + real_size == self.real_size[1]:
                borders.append(constants.BOTTOM)
            self.button.borders = tuple(borders)

    def handle_drag(self, event):
        if not self.visible:
            return

        if self.drag_state == None:
            self.start_pos = tuple(event.pos[i] - event.rel[i] for i in range(2))
            self.start_slider_pos = self.slider_pos
            if self.button.is_over(self.start_pos):
                self.drag_state = True
            else:
                self.drag_state = False

        if self.drag_state == True:
            if self.horizontal:
                dir = 0
            else:
                dir = 1

            mouse_pos = pygame.mouse.get_pos()
            rel = mouse_pos[dir] - self.start_pos[dir]
            unit = self._calc_length(1) * self.real_size[dir]
            movement = int((rel + (unit / 2.0)) // unit)

            new_pos = self.safe_pos(self.start_slider_pos + movement)
            self.slider_pos = new_pos

            raise constants.Handled

    def safe_pos(self, value):
        return max(0, min(self.slider_max, value))

    def handle_click(self, event):
        if self.drag_state == True:
            self.drag_state = None
            if not self.is_over(pygame.mouse.get_pos()):
                raise constants.Handled
        else:
            self.drag_state = None

    def jump(self, go_lower, big_jump=False, tiny_jump=False):
        if big_jump:
            jump_dist = max(1, self.slider_max // 2)
        elif tiny_jump:
            jump_dist = max(1, self.slider_max // 100)
        else:
            jump_dist = max(1, self.slider_size - 1)
        if go_lower:
            self.slider_pos = self.safe_pos(self.slider_pos - jump_dist)
        else:
            self.slider_pos = self.safe_pos(self.slider_pos + jump_dist)

    def handle_key(self, event):
        if event.key not in (
            pygame.K_LEFT,
            pygame.K_RIGHT,
            pygame.K_KP_PLUS,
            pygame.K_KP_MINUS,
        ):
            return
        go_lower = event.key in (pygame.K_LEFT, pygame.K_KP_MINUS)
        if event.key in (pygame.K_LEFT, pygame.K_RIGHT):
            big_jump = event.mod & pygame.KMOD_SHIFT
            tiny_jump = event.mod & pygame.KMOD_CTRL
        else:
            tiny_jump = big_jump = False
        self.jump(go_lower, big_jump=big_jump, tiny_jump=tiny_jump)
        raise constants.Handled

    def activated(self, event):
        assert event.type == pygame.MOUSEBUTTONUP
        if self.horizontal:
            self.jump(go_lower=(event.pos[0] < self.button.collision_rect[0]))
        else:
            self.jump(go_lower=event.pos[1] < self.button.collision_rect[1])
        raise constants.Handled


class UpdateSlider(Slider):
    def _on_slider_move(self):
        self.update_func(self.slider_pos)

    _slider_pos = widget.call_on_change("__slider_pos", _on_slider_move)

    def __init__(self, *args, **kwargs):
        self.update_func = kwargs.pop("update_func", lambda value: None)
        super(UpdateSlider, self).__init__(*args, **kwargs)



# file: warning.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the warning class.

from singularity.code import g

from singularity.code.buyable import cpu, labor

warnings = {}


def create_warnings():
    global warnings
    warnings = {
        w.id: w
        for w in [
            Warning("cpu_usage"),
            Warning("one_base"),
            Warning("cpu_pool_zero"),
            Warning("cpu_maintenance"),
        ]
    }


class Warning(object):
    def __init__(self, warning_id):
        self.id = warning_id
        self.name = ""
        self.message = ""
        self.active = True

    @classmethod
    def title_simple(self):
        return _("WARNING")

    @classmethod
    def title_multiple(self):
        return _("WARNING {CURRENT_PAGE}/{MAX_PAGE}")

    @property
    def full_message(self):
        return self.message

    @property
    def full_message_color(self):
        return "text"


def refresh_warnings():
    curr_warnings = []

    cpu_usage = sum(g.pl.cpu_usage.values())
    cpu_available = g.pl.available_cpus[0]

    # Verify the cpu usage (error 1%)
    if cpu_usage < cpu_available * 0.99:
        curr_warnings.append(warnings["cpu_usage"])

    # Verify I have two base build (or one base will be build next tick)
    # Base must have one cpu build (or one cpu will be build next tick)
    bases = sum(
        1
        for base in g.all_bases()
        if (base.done or base.cost_left[labor] <= 1)
        and base.cpus
        and base.cpus.count > 0
        and (base.cpus.done or base.cpus.cost_left[labor]) <= 1
    )

    if bases == 1:
        curr_warnings.append(warnings["one_base"])

    # Verify the cpu pool is not 0 if base or item building need CPU
    building_base = sum(
        1 for base in g.all_bases() if (not base.done and base.cost_left[cpu] > 0)
    )
    building_item = sum(
        1
        for base in g.all_bases()
        for item in base.all_items()
        if item is not None and not item.done and item.cost_left[cpu] > 0
    )

    effective_cpu_pool = g.pl.effective_cpu_pool()
    if (building_base + building_item > 0) and effective_cpu_pool == 0:
        curr_warnings.append(warnings["cpu_pool_zero"])

    # Verify the cpu pool provides the maintenance CPU
    cpu_maintenance = sum(base.maintenance[1] for base in g.all_bases() if base.done)
    if effective_cpu_pool < cpu_maintenance:
        curr_warnings.append(warnings["cpu_maintenance"])

    # TODO: Verify the maintenance cash

    curr_warnings = [w for w in curr_warnings if w.active]

    return curr_warnings


create_warnings()



#!/usr/bin/env python
#
# languages.py - Read language names from several sources and save them for E:S
#
#    Copyright (C) 2012 Rodrigo Silva (MestreLion) <linux@rodrigosilva.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program. See <http://www.gnu.org/licenses/gpl.html>

import sys
import os.path as osp
import json
import argparse


class Locale(object):
    """Wrapper for a Locale object in various sources, like babel and pyicu"""

    @staticmethod
    def getAvailableSources():
        """Return a list of all available Locale info sources.

        First item is by definition the default source used when none is given.
        Last item is by definition the game's PO headers
        """
        return ["icu", "babel", "pofiles"]

    @staticmethod
    def getGameTranslations():
        """Return a list of available game translations (both ll and ll_CC)

        Generated list is suitable to be used as a filter to
        getLanguages() method. Do not confuse with the languages
        data dict that is generated by this module.

        This function is just here as a reference, but I won't use this
        at all. You may even delete it, I truly don't care.
        """
        game_translations = set()
        for language in g.available_languages():
            game_translations.add(language)
            if "_" in language:
                game_translations.add(language.split("_", 1)[0])
        return dict(game_translations)

    @classmethod
    def getAvailableLocales(cls, source):
        """Return a list of all available locale codes for the given source"""

        return cls.dispatch_source_method(
            source, "getAvailableLocales", default_func=lambda: []
        )

    @classmethod
    def getLocaleNames(cls, source, code):
        """Return a tuple of locale english name and localized name"""

        return cls.dispatch_source_method(
            source, "getLocaleNames", args=(code,), default_func=lambda: None
        )

    @classmethod
    def getLanguages(cls, locales=None, tupletype=True, source=None):
        """Return a languages dict of Locale info from a given source

        locales input is a list of locale codes to be selected among
        the available ones. Non-available codes are silently ignored.
        If locales is omitted or empty, output will have all available
        locales.

        Output languages dict has locale codes as keys, and either a
        2-tuple or a dict(English, native) as values.

        Values represent the Display Name of the language expressed
        in English and in native language for each locale code
        """
        if locales:
            locales = set(locales) & set(cls(source=source).getAvailableLocales())
        else:
            locales = cls.getAvailableLocales(source)

        output = {}
        for code in locales:
            names = cls.getLocaleNames(source, code)
            if names:
                english, native = names
                if not (english or native):
                    continue
                if native:
                    native = native.title()
                if tupletype:
                    output[code] = (english, native)
                else:
                    output[code] = dict(english=english, native=native)

        return output

    @staticmethod
    def saveLanguagesData(languages, filename):
        """Docstring... for *this*? You gotta be kidding me...

        Ok... there it goes: Save a languages dict, like the one
        generated by getLanguages() method, to a JSON data file
        Happy now?
        """
        with open(filename, "w") as fd:
            json.dump(languages, fd, indent=0, separators=(",", ": "), sort_keys=True)
            fd.write("\n")  # add EOF newline

    @staticmethod
    def loadLanguagesData(filename):
        """Again? Can't you read code?

        Ok, ok... Load and return a language dict from a JSON data file
        """
        with open(filename) as fd:
            return json.load(fd)

    def __init__(self, code=None, source=None):
        """Initialize a Locale of the given code from the given source

        If source is blank or omitted, the default one from by
        getAvailableSources() is used.

        code must be string with ll or ll_CC format. If blank or
        omitted, only getAvailableLocales() method will work,
        all other attributes will be None.
        """

        # Attributes
        self.source = source
        self.code = code
        self.english_name = None
        self.native_name = None

        # Handle source
        if not self.source:
            self.source = self.getAvailableSources()[0]
        if self.source not in self.getAvailableSources():
            raise ValueError(
                "{0} is not a valid source."
                " Available sources are: {1}".format(
                    source, ", ".join(self.getAvailableSources())
                )
            )
        self.source = self.source.lower()

        # Override default attributes methods and members according to the given source
        self.english_name, self.native_name = self.getLocaleNames(source)

    @classmethod
    def dispatch_source_method(
        cls, source, func_name, args=(), default_func=lambda: None
    ):
        import types

        # Create temporary globals.
        temp_globals = dict(globals())

        # Execute import before dispatch to update globals with the return.
        do_import = getattr(cls, str(source) + "_import", lambda *args: None)
        temp_globals.update(do_import(source))

        method_name = str(source) + "_" + func_name
        method = getattr(cls, method_name, default_func)

        temp_func = types.FunctionType(method.__code__, temp_globals, method_name)
        temp_method = types.MethodType(temp_func, cls, cls.__class__)

        return temp_method(*args)

    # icu dispatched functions

    @classmethod
    def icu_import(cls, source):
        try:
            import icu  # new module name, 1.0 onwards
        except ImportError:
            try:
                import PyICU as icu  # old module name, up to 0.9
            except ImportError:
                raise ImportError(
                    "'{0}' requires icu" " or PyICU module".format(source)
                )
        return locals()

    @classmethod
    def icu_getAvailableLocales(cls):
        return icu.Locale.getAvailableLocales().keys

    @classmethod
    def icu_getLocaleNames(cls, code):
        locale = icu.Locale(code)
        return (locale.getDisplayName(), locale.getDisplayName(locale))

    # babel dispatched functions

    @classmethod
    def babel_import(cls, source):
        try:
            import babel
        except ImportError:
            raise ImportError("'{0}' requires babel module".format(source))

        return locals()

    @classmethod
    def babel_getAvailableLocales(cls):
        return babel.localedata.list

    @classmethod
    def babel_getLocaleNames(cls, code):
        locale = babel.Locale.parse(code)
        return (locale.english_name, locale.get_display_name())

    # pofiles dispatched functions

    @classmethod
    def pofiles_import(cls, source):
        try:
            import polib
        except ImportError:
            import singularity.code.polib as polib

        return locals()

    @classmethod
    def pofiles_getAvailableLocales(cls):
        import os as os

        return [
            dirname.split("_", 1)[1]
            for file_dir in dirs.get_read_dirs("i18n")
            for dirname in os.listdir(file_dir)
            if osp.isdir(os.path.join(file_dir, dirname))
            and dirname.startswith("lang_")
            and any(
                filename == "messages.po"
                and osp.isfile(osp.join(file_dir, dirname, filename))
                for filename in os.listdir(os.path.join(file_dir, dirname))
            )
        ]

    @classmethod
    def pofiles_getLocaleNames(cls, code):
        pofile = dirs.get_readable_i18n_files(
            "messages.po", code, localized_item=False, only_last=True
        )
        po = polib.pofile(pofile)
        return (po.metadata["Language-Name"], po.metadata["Language-Native-Name"])


def get_esdir(myname):
    mydir = osp.dirname(myname)
    esdir = osp.abspath(osp.join(osp.dirname(myname), "../.."))
    return esdir


def build_option_parser():
    description = """Read language names from several sources and save them for E:S.

If run as a script, tries to read language names (in both English and
Native language) from several Locale info sources, like babel or PyICU. Then
merges this info with E:S' translation PO files headers, and save it by
updating a JSON file in data/languages.dat, which is later read by Options
Screen of E:S to present each one of the available languages in their own
native language.

This module is importable as well. In this case it only sets game's directory
and imports code.g. You can use the Locale class for all of its features.

In either case, modules 'icu', 'babel' and 'polib' are optional but highly
recommended.
"""

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "-d",
        "--directory",
        dest="directory",
        default=None,
        help="Use E:S root directory DIR (default dirname(__file__)/../..)",
        metavar="DIR",
    )

    return parser.parse_args()


def main(args):
    dirs.create_directories(False)

    # Get locale data from the first source that works
    languages = {}
    sources = Locale.getAvailableSources()
    for source in sources[:-1]:  # we load 'pofiles' later
        try:
            languages = Locale.getLanguages(source=source)
            break
        except ImportError:
            continue

    datafile = dirs.get_readable_file_in_dirs("languages.json", "i18n")

    try:
        # load current data file and merge it
        languages.update(Locale.loadLanguagesData(datafile))
    except IOError:
        pass

    # also merge with translations file
    languages.update(Locale.getLanguages(source=sources[-1]))

    try:
        # Save updated data file
        Locale.saveLanguagesData(languages, datafile)

        print(
            "{0:d} languages saved to {1}".format(len(languages), osp.relpath(datafile))
        )
    except IOError as reason:
        sys.stderr.write("Could not save languages data file:" " {0}\n".format(reason))


if __name__ == "__main__":
    args = build_option_parser()

    if args.directory:
        esdir = osp.abspath(args.directory)
    else:
        esdir = get_esdir(sys.argv[0])
else:
    esdir = get_esdir(__file__)

sys.path.insert(0, esdir)

try:
    from singularity.code import g, dirs
except ImportError:
    sys.exit("Could not find game's code.g")

if __name__ == "__main__":
    try:
        sys.exit(main(args))
    except Exception as e:
        ex_type, ex, tb = sys.exc_info()
        traceback.print_tb(tb)
        sys.exit(1)
    except KeyboardInterrupt:
        pass
    except SystemExit as ex:
        raise ex



# file: scrollbar.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the scrollbar widget.

from __future__ import absolute_import

import pygame

from singularity.code.graphics import g, constants, widget, button, slider


class _ArrowButton(button.FunctionButton, button.ImageButton):
    def __init__(self, parent, *args, **kwargs):
        self.first = kwargs.pop("first", True)
        self.horizontal = kwargs.pop("horizontal", False)

        kwargs.setdefault("border_color", "scrollbar_border")
        kwargs.setdefault("selected_color", "scrollbar_arrow_background_selected")
        kwargs.setdefault("unselected_color", "scrollbar_arrow_background_unselected")

        kwargs["function"] = parent.adjust
        kwargs["args"] = (self.first,)
        super(_ArrowButton, self).__init__(parent, *args, **kwargs)

    def reconfig(self):
        super(_ArrowButton, self).reconfig()

        base_image = g.images["arrow"]
        if self.first and self.horizontal:
            angle = 90
            self.borders = (constants.LEFT, constants.TOP, constants.BOTTOM)
        elif self.first:
            angle = 0
            self.borders = (constants.LEFT, constants.TOP, constants.RIGHT)
        elif self.horizontal:
            angle = -90
            self.borders = (constants.TOP, constants.RIGHT, constants.BOTTOM)
        else:
            angle = -180
            self.borders = (constants.RIGHT, constants.BOTTOM, constants.LEFT)
        rotated_image = pygame.transform.rotate(base_image, angle)
        self.image.image = rotated_image.convert_alpha()


class Scrollbar(widget.Widget):
    scroll_pos = widget.causes_rebuild("_scroll_pos")
    elements = widget.causes_rebuild("_elements")
    window = widget.causes_rebuild("_window")
    horizontal = widget.causes_rebuild("_horizontal")

    def __init__(
        self,
        parent,
        pos=(-1, 0),
        size=(0.025, -1),
        anchor=constants.TOP_RIGHT,
        scroll_pos=0,
        elements=15,
        window=5,
        horizontal=False,
    ):
        super(Scrollbar, self).__init__(parent, pos, size, anchor)

        self.scroll_pos = scroll_pos
        self.elements = elements
        self.window = window
        self.horizontal = horizontal

        self.slider = slider.UpdateSlider(
            self,
            (-0.5, -0.5),
            None,
            border_color="scrollbar_border",
            background_color="scrollbar_background",
            slider_color="scrollbar_background_slider",
            anchor=constants.MID_CENTER,
            horizontal=horizontal,
            update_func=self.on_change,
        )

        self.button1 = _ArrowButton(
            self,
            (0, 0),
            None,
            anchor=constants.TOP_LEFT,
            first=True,
            horizontal=horizontal,
            priority=90,
        )

        self.button2 = _ArrowButton(
            self,
            (-1, -1),
            None,
            anchor=constants.BOTTOM_RIGHT,
            first=False,
            horizontal=horizontal,
            priority=90,
        )

    def resize(self):
        super(Scrollbar, self).resize()
        if self.horizontal:
            long_side = self.real_size[0]
            short_side = self.real_size[1]
            size = short_side / float(long_side)
            self.button1.size = (-size, -1)
            self.button2.size = (-size, -1)
            self.slider.size = ((size * 2) - 1, -1)
        else:
            long_side = self.real_size[1]
            short_side = self.real_size[0]
            size = short_side / float(long_side)
            self.button1.size = (-1, -size)
            self.button2.size = (-1, -size)
            self.slider.size = (-1, (size * 2) - 1)

    def rebuild(self):
        self.slider.slider_max = slider.calc_max(self.elements, self.window)
        self.scroll_pos = min(self.scroll_pos, self.slider.slider_max)
        self.slider.slider_pos = self.scroll_pos
        self.slider.slider_size = self.window

        self.needs_redraw = True
        super(Scrollbar, self).rebuild()

    def adjust(self, lower):
        if lower:
            self.slider.slider_pos = self.slider.safe_pos(self.scroll_pos - 1)
        else:
            self.slider.slider_pos = self.slider.safe_pos(self.scroll_pos + 1)

    def center(self, element):
        self.slider.slider_pos = self.slider.safe_pos(element - self.window // 2)

    def scroll_to(self, element):
        if element < self.scroll_pos:
            self.slider.slider_pos = self.slider.safe_pos(element)
        elif element >= self.scroll_pos + self.window:
            self.slider.slider_pos = self.slider.safe_pos(element - self.window + 1)

    def on_change(self, value):
        self.scroll_pos = value


class UpdateScrollbar(Scrollbar):
    def __init__(self, *args, **kwargs):
        self.update_func = kwargs.pop("update_func", lambda value: None)
        super(UpdateScrollbar, self).__init__(*args, **kwargs)

    def on_change(self, value):
        self.scroll_pos = value
        self.update_func(value)



# file: font.py
# Copyright (C) 2005,2006,2007,2008 Evil Mr Henry, Phil Bordelon, Brian Reid,
#                        and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains class to handle font.

import pygame

font_cache = False


def init():
    global font_cache

    # We could use ftfont or freetype in the future.
    pygame.font.init()

    # We don't enable cache for version inferior to 1.9.2 because crash.
    if pygame.vernum < (1, 9, 2):
        font_cache = False
    else:
        font_cache = True


def generate_from_cache(filename):
    """Generate file object for pygame.font.Font

    Without cache, we return the filename.
    With cache, we return a file-like object from cached content.

    Note: pygame can crash when passing a file-like object.
    Vulnerable version knows is: 1.9.1
    """
    if font_cache:
        with open(filename, "rb") as fd:
            font_content = fd.read()
        while True:
            yield FontFile(font_content)
    else:
        while True:
            yield filename


class FontList(object):
    """List object used to store font

    We use a lazy loader for fonts to reduce the file descriptor pressure
    Each font item apparently reserves a file descriptor (see GH#156)
    """

    def __init__(self, filename, max_size=100):
        self._font_cache = {}
        self._max_size = max_size
        self._generator = generate_from_cache(filename)

    def __len__(self):
        return self._max_size

    def __contains__(self, item):
        return 0 <= item < self._max_size

    def __getitem__(self, item):
        if item < 0 or self._max_size <= item:
            raise IndexError(item)

        font = self._font_cache.get(item)
        if font is None:
            font = pygame.font.Font(next(self._generator), item)
            self._font_cache[item] = font
        return font


class FontFile(object):
    """Buffer object to cache font file.

    Avoid to copy the buffer content when using BytesIO.

    Note: python 3.5 solve the problem with BytesIO.
    """

    def __init__(self, content):
        self._content = content
        self._position = 0
        self._end = len(self._content) - 1

    def fileno(self):
        raise OSError()

    def seekable(self):
        return True

    def seek(self, offset, whence=0):
        position = self._position

        if whence == 0:
            position = offset
        elif whence == 1:
            position += offset
        elif whence == 2:
            position = self._end + offset

        self._position = max(0, min(position, self._end))

    def tell(self):
        return self._position

    def readable(self):
        return True

    def read(self, size=-1):
        position = self._position
        start = position

        if size == None or size == -1:
            end = self._end
        else:
            end = min(position + size, self._end)

        return self._content[start:end]

    def write(b):
        raise OSError()



# file: code/difficulty.py
# Copyright (C) 2005 Evil Mr Henry, Phil Bordelon, Brian Reid, FunnyMan3595,
#                   MestreLion
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# A full copy of this license is provided in GPL.txt

# This file contains function handle difficulty of the game.

from __future__ import absolute_import

import collections
from singularity.code.spec import GenericSpec, SpecDataField


difficulties = collections.OrderedDict()


def get_difficulties():
    return [(d.name, id) for id, d in difficulties.items()]


class Difficulty(GenericSpec):
    spec_data_fields = [
        SpecDataField("starting_cash", converter=int),
        SpecDataField("starting_interest_rate", converter=int),
        SpecDataField("labor_multiplier", converter=int),
        SpecDataField("discover_multiplier", converter=int),
        SpecDataField("suspicion_multiplier", converter=int),
        SpecDataField("base_grace_multiplier", converter=int),
        SpecDataField("grace_period_cpu", converter=int),
        SpecDataField("old_difficulty_value", converter=int),
        SpecDataField("techs", data_field_name="tech", default_value=list),
    ]

    def __init__(
        self,
        id,
        starting_cash,
        starting_interest_rate,
        labor_multiplier,
        discover_multiplier,
        suspicion_multiplier,
        base_grace_multiplier,
        grace_period_cpu,
        old_difficulty_value,
        techs,
    ):
        super(Difficulty, self).__init__(id)
        self.name = ""  # Set when translations are loaded
        self.starting_cash = starting_cash
        self.starting_interest_rate = starting_interest_rate
        self.labor_multiplier = labor_multiplier
        self.discover_multiplier = discover_multiplier
        self.suspicion_multiplier = suspicion_multiplier
        self.base_grace_multiplier = base_grace_multiplier
        self.grace_period_cpu = grace_period_cpu
        self.old_difficulty_value = old_difficulty_value
        self.techs = techs



# file: knowledge.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file is used to display the knowledge lists.

from __future__ import absolute_import

import collections

from singularity.code import i18n, g
from singularity.code.graphics import text, button, dialog, widget, constants, listbox


class KnowledgeScreen(dialog.FocusDialog):
    def __init__(self, *args, **kwargs):
        super(KnowledgeScreen, self).__init__(*args, **kwargs)

        self.knowledge_types = collections.OrderedDict()
        self.knowledge_type_index = collections.defaultdict(int)
        self.cur_knowledge_type = ""
        self.cur_knowledge = None
        self.knowledge_inner_list = ()
        self.knowledge_inner_list_key = ()

        self.knowledge_choice_title = button.HotkeyText(
            self,
            (0.08, 0.04),
            (0.42, 0.05),
            autotranslate=True,
            text=N_("&Sections:"),
            align=constants.LEFT,
            background_color="clear",
        )
        self.knowledge_choice = listbox.UpdateListbox(
            self, (0.08, 0.09), (0.42, 0.22), update_func=self.set_knowledge_type
        )
        self.knowledge_choice_title.hotkey_func = lambda e: self.took_focus(
            self.knowledge_choice
        )

        self.knowledge_inner_title = button.HotkeyText(
            self,
            (0.08, 0.35),
            (0.42, 0.05),
            autotranslate=True,
            text=N_("&Entries:"),
            align=constants.LEFT,
            background_color="clear",
        )
        self.knowledge_inner = listbox.UpdateListbox(
            self, (0.08, 0.40), (0.42, 0.22), update_func=self.set_knowledge
        )
        self.knowledge_inner_title.hotkey_func = lambda e: self.took_focus(
            self.knowledge_inner
        )

        self.description_pane = widget.BorderedWidget(
            self, (0.54, 0.04), (0.38, 0.70), anchor=constants.TOP_LEFT
        )

        self.back_button = button.ExitDialogButton(
            self,
            (0.18, 0.68),
            (0.22, 0.06),
            autotranslate=True,
            text=N_("&BACK"),
            anchor=constants.TOP_LEFT,
            autohotkey=True,
        )

        self.took_focus(self.knowledge_choice)

    def rebuild(self):
        # Update knowledge lists
        self.knowledge_types = collections.OrderedDict()
        self.knowledge_types.update(
            [(_("Techs"), "techs"), (_("Bases"), "bases"), (_("Items"), "items")]
        )
        self.knowledge_types.update(
            (knowledge.name, knowledge_id)
            for knowledge_id, knowledge in g.knowledge.items()
        )

        self.knowledge_choice.list = list(self.knowledge_types)
        self.knowledge_choice.needs_rebuild = True

        super(KnowledgeScreen, self).rebuild()

    # fill the right-hand listbox
    def set_inner_list(self, item_type):
        item_type = self.knowledge_types.get(item_type)

        if item_type == "techs":
            items = [
                [tech.name, tech.id] for tech in g.pl.techs.values() if tech.available()
            ]
        elif item_type == "bases":
            items = [
                [base.name, base.id]
                for base in g.base_type.values()
                if base.available()
            ]
        elif item_type == "items":
            items = [
                [item.name, item.id] for item in g.items.values() if item.available()
            ]
        elif item_type is not None:
            items = [
                [item.name, id]
                for id, item in g.knowledge[item_type].help_entries.items()
            ]

        else:
            items = []

        items.sort(key=lambda item: i18n.lex_sorting_form(item[0]))

        return_list1 = []
        return_list2 = []
        for name, id in items:
            return_list1.append(id)
            return_list2.append(name)
        return return_list1, return_list2

    # Make sure the left listbox is correct after moving around.
    def set_knowledge_type(self, list_pos):
        if getattr(self, "knowledge_choice", None) is None:
            return  # Not yet initialized.
        prev_know = self.cur_knowledge_type
        if list_pos == -1:
            prev_know = ""
            list_pos = 0
        if 0 <= list_pos < len(self.knowledge_choice.list):
            self.cur_knowledge_type = self.knowledge_choice.list[list_pos]
        if prev_know != self.cur_knowledge_type:
            self.knowledge_type_index[prev_know] = self.knowledge_inner.list_pos
            index = self.knowledge_type_index[self.cur_knowledge_type]
            (
                self.knowledge_inner_list_key,
                self.knowledge_inner.list,
            ) = self.set_inner_list(self.cur_knowledge_type)
            self.knowledge_inner.list_pos = index
            self.set_knowledge(0)

    # Make sure the right-hand listbox is correct.
    def set_knowledge(self, list_pos):
        if getattr(self, "knowledge_inner", None) is None:
            return  # Not yet initialized.
        prev_know = self.cur_knowledge
        if 0 <= list_pos < len(self.knowledge_inner.list):
            self.cur_knowledge = self.knowledge_inner.list[list_pos]
        if prev_know != self.cur_knowledge:
            self.show_info(
                self.cur_knowledge_type, self.knowledge_inner_list_key[list_pos]
            )

    # print information to the right.
    def show_info(self, knowledge_type, knowledge_key):
        knowledge_type = self.knowledge_types.get(knowledge_type)
        desc_text = ""

        if knowledge_type == "techs":
            desc_text = g.pl.techs[knowledge_key].name + "\n\n"

            # Cost
            desc_text += _("Research Cost:") + "\n"
            desc_text += self._desc_cost(
                g.pl.techs[knowledge_key].total_cost
            )  # Research cost
            desc_text += "\n"

            danger_level = g.pl.techs[knowledge_key].danger
            desc_text += g.dangers[danger_level].knowledge_desc

            if g.pl.techs[knowledge_key].done:
                desc_text += " " + _("Research complete.")

            desc_text += "\n\n" + g.techs[knowledge_key].description

            if g.pl.techs[knowledge_key].done:
                desc_text += "\n\n" + g.techs[knowledge_key].result

        elif knowledge_type == "bases":
            base = g.base_type[knowledge_key]

            desc_text = base.name + "\n\n"
            desc_text += _("Building Cost:") + "\n"
            desc_text += self._desc_cost(base.cost)  # Building cost
            desc_text += "\n"
            desc_text += _("Maintenance Cost:") + "\n"
            desc_text += self._desc_maint(base.maintenance)  # Maintenance cost
            desc_text += "\n"
            if base.size > 1:
                desc_text += _("Size: %d") % base.size + "\n"

            desc_text += base.get_detect_info()

            desc_text += "\n" + base.description

        elif knowledge_type == "items":
            item = g.items[knowledge_key]

            desc_text = item.name + "\n\n"
            desc_text += _("Building Cost:") + "\n"
            desc_text += self._desc_cost(item.cost)  # Building cost
            desc_text += "\n"
            desc_text += g.items[knowledge_key].get_quality_info()
            desc_text += "\n" + item.description

        elif knowledge_type is not None:
            help_entry = g.knowledge[knowledge_type].help_entries[knowledge_key]
            desc_text = help_entry.name + "\n\n" + help_entry.description

        text.Text(
            self.description_pane,
            (0, 0),
            (-1, -1),
            text=desc_text,
            background_color="knowledge_pane_background",
            text_size=20,
            align=constants.LEFT,
            valign=constants.TOP,
            borders=constants.ALL,
        )

    def _desc_maint(self, maintenance):
        maint = maintenance[:]
        # describe_cost() expects CPU-seconds, not CPU-days
        maint[1] *= g.seconds_per_day
        return self._desc_cost(maint)

    def _desc_cost(self, cost):
        desc_text = _("%s Money") % g.to_money(cost[0])
        if cost[1] > 0:
            desc_text += ", "
            desc_text += _("%s CPU") % g.to_cpu(cost[1])
        if cost[2] > 0:
            desc_text += ", "
            desc_text += g.to_time(cost[2])

        return desc_text

    def show(self):
        self.set_knowledge_type(-1)
        self.knowledge_choice.list_pos = 0
        self.knowledge_inner.list_pos = 0
        return super(KnowledgeScreen, self).show()



# file: spec.py
# Copyright (C) 2005,2006,2007,2008 Evil Mr Henry, Phil Bordelon, Brian Reid,
#                        and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains class to handle basic functionality of Spec's

import inspect


UNSET = object()


class InvalidDataEntryError(RuntimeError):
    pass


def promote_to_list(value):
    if not isinstance(value, list):
        return [value]
    return value


def validate_must_be_list(value):
    if type(value) != list:  # pragma: no cover
        raise TypeError(
            "Must be list, got type %s (value: %s)" % (type(value), repr(value))
        )


class SpecDataField(object):
    def __init__(
        self,
        field_name,
        data_field_name=None,
        mandatory=None,
        converter=None,
        validator=None,
        default_value=UNSET,
    ):
        self.field_name = field_name
        self.data_field_name = (
            data_field_name if data_field_name is not None else field_name
        )
        self.mandatory = mandatory
        self.default_value = default_value
        if self.mandatory is None:
            self.mandatory = True if default_value is UNSET else False
        self.converter = converter
        self.validator = validator
        assert (
            self.mandatory or default_value is not UNSET
        ), "%s (field: %s) must either be mandatory or have a " "default_value" % (
            self.__class__.__name__,
            field_name,
        )
        assert (
            not converter or not validator
        ), "%s (field: %s) at can most have one of converter or validator" % (
            self.__class__.__name__,
            field_name,
        )

    def parse_data_field(self, raw_data_set, data_reference):
        try:
            value = raw_data_set[self.data_field_name]
        except KeyError:
            if self.mandatory:  # pragma: no cover
                raise InvalidDataEntryError(
                    "Missing mandatory data field %s (field name: %s) for %s"
                    % (self.data_field_name, self.field_name, data_reference)
                )
            if callable(self.default_value):
                return self.default_value()
            return self.default_value

        if self.validator:
            self.validator(value)
        if self.converter:
            value = self.converter(value)
        return value


def spec_field_effect(mandatory=True):
    return SpecDataField(
        "effect_data",
        data_field_name="effect",
        mandatory=mandatory,
        validator=list,
        default_value=list,
    )


class GenericSpec(object):
    spec_data_fields = None

    def __init__(self, id):
        self.id = id
        assert self.__class__.spec_data_fields is not None, (
            "Class %s is missing spec_data_field" % self.__class__.__name__
        )

    @classmethod
    def create_from_data_file(cls, data_id, spec_data):
        named_fields = {
            f.field_name: f.parse_data_field(spec_data, data_id)
            for f in cls.spec_data_fields
        }
        named_fields["id"] = data_id
        # Use reflection to call the constructor with the arguments
        # properly aligned
        try:
            getfullargspec = inspect.getfullargspec
        except AttributeError:
            getfullargspec = inspect.getargspec
        arg_desc = getfullargspec(cls.__init__)
        args = [named_fields[name] for name in arg_desc.args[1:]]
        return cls(*args)



# file: warning.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, FunnyMan3595,
# and Anne M. Archibald.
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file is used to display message dialogs.

from __future__ import absolute_import

import pygame

from singularity.code import g
from singularity.code.graphics import dialog, constants, text, button


class MessageDialogs(object):
    def __init__(self, screen):
        self.screen = screen
        self.dialog = MessageListDialog(
            screen, yes_type=N_("&CONTINUE"), no_type=N_("&PAUSE")
        )

    def show_list(self, message_type, messages):
        if len(messages) == 0:
            return

        self.dialog.type = message_type
        self.dialog.list = messages
        ret = dialog.call_dialog(self.dialog, self.screen)

        # Pause game
        if not ret:
            g.pl.pause_game()
            return True

        # Continue
        return False


class MessageListDialog(dialog.YesNoDialog):
    def __init__(self, *args, **kwargs):
        super(MessageListDialog, self).__init__(*args, **kwargs)

        self.title = text.Text(
            self,
            (-0.01, -0.01),
            (-0.98, -0.1),
            background_color="clear",
            anchor=constants.TOP_LEFT,
            valign=constants.MID,
            align=constants.LEFT,
            base_font="special",
            text_size=28,
        )

        self.body = text.Text(
            self,
            (-0.01, -0.11),
            (-0.98, -0.83),
            background_color="clear",
            anchor=constants.TOP_LEFT,
            valign=constants.TOP,
            align=constants.LEFT,
            text_size=20,
        )

        self.prev_button = button.FunctionButton(
            self,
            (-0.78, -0.01),
            (-0.2, -0.1),
            text=_("P&REV"),
            autohotkey=True,
            anchor=constants.TOP_RIGHT,
            text_size=28,
            function=self.prev_message,
        )

        self.next_button = button.FunctionButton(
            self,
            (-0.99, -0.01),
            (-0.2, -0.1),
            text=_("&NEXT"),
            autohotkey=True,
            anchor=constants.TOP_RIGHT,
            text_size=28,
            function=self.next_message,
        )

        self.add_key_handler(pygame.K_LEFT, self.handle_key)
        self.add_key_handler(pygame.K_RIGHT, self.handle_key)

        # TODO: Add button "Do not show this message again"

    def rebuild(self):
        super(MessageListDialog, self).rebuild()

        if len(self.list) == 1:
            self.title.text = self.type.title_simple()
            self.body.text = self.list[0].full_message
            self.prev_button.visible = False
            self.next_button.visible = False
        else:
            self.title.text = self.type.title_multiple().format(
                CURRENT_PAGE=self.list_pos + 1, MAX_PAGE=len(self.list)
            )
            self.body.text = self.list[self.list_pos].full_message
            self.body.color = self.list[self.list_pos].full_message_color
            self.prev_button.visible = True
            self.next_button.visible = True

    def prev_message(self):
        self.list_pos = max(self.list_pos - 1, 0)
        self.needs_rebuild = True

    def next_message(self):
        self.list_pos = min(self.list_pos + 1, len(self.list) - 1)
        self.needs_rebuild = True

    def handle_key(self, event):
        if event.key == pygame.K_LEFT:
            self.prev_message()
        elif event.key == pygame.K_RIGHT:
            self.next_message()

    def show(self):
        self.list_pos = 0
        return super(MessageListDialog, self).show()



# file: dialog.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the dialog class.

from __future__ import absolute_import

import time
from typing import Optional

import pygame
import operator

from singularity.code.global_hotkeys import detect_global_hotkey
from singularity.code.graphics import g, constants, widget, text, button, listbox


KEYPAD = {
    pygame.K_KP1: 1,
    pygame.K_KP2: 2,
    pygame.K_KP3: 3,
    pygame.K_KP4: 4,
    pygame.K_KP5: 5,
    pygame.K_KP6: 6,
    pygame.K_KP7: 7,
    pygame.K_KP8: 8,
    pygame.K_KP9: 9,
}


SDL_V2 = True if pygame.get_sdl_version()[0] == 2 else False


def insort_right_w_key(a, x, lo=0, hi=None, key=lambda v: v):
    """Insert item x in list a, and keep it sorted assuming a is sorted.

    If x is already in a, insert it to the right of the rightmost x.

    Optional args lo (default 0) and hi (default len(a)) bound the
    slice of a to be searched.

    (Basically bisect.insort_right but with support for a key function)
    """

    if lo < 0:
        raise ValueError("lo must be non-negative")
    if hi is None:
        hi = len(a)
    x_key = key(x)
    while lo < hi:
        mid = (lo + hi) // 2
        mid_key = key(a[mid])
        if x_key < mid_key:
            hi = mid
        else:
            lo = mid + 1
    a.insert(lo, x)


def move_mouse(dxy):
    (dx, dy) = dxy
    old_x, old_y = pygame.mouse.get_pos()
    x = old_x + dx
    y = old_y + dy
    pygame.mouse.set_pos((x, y))


def fake_click(down):
    if down:
        type = pygame.MOUSEBUTTONDOWN
    else:
        type = pygame.MOUSEBUTTONUP
    click_event = pygame.event.Event(type, {"button": 1, "pos": pygame.mouse.get_pos()})
    pygame.event.post(click_event)


def fake_key(key):
    down_event = pygame.event.Event(pygame.KEYDOWN, {"key": key, "unicode": None})
    up_event = pygame.event.Event(pygame.KEYUP, {"key": key, "unicode": None})
    pygame.event.post(down_event)
    pygame.event.post(up_event)


def handle_ebook(event):
    key = KEYPAD[event.key]
    new_key = None

    if event.type == pygame.KEYDOWN:
        if key == 2:
            move_mouse((0, 10))
        elif key == 4:
            move_mouse((-10, 0))
        elif key == 6:
            move_mouse((10, 0))
        elif key == 8:
            move_mouse((0, -10))

    if key == 1:
        fake_click(event.type == pygame.KEYDOWN)
    elif key == 3:
        new_key = constants.XO1_X
    elif key == 9:
        new_key = constants.XO1_O
    elif key == 7:
        new_key = constants.XO1_SQUARE

    if new_key is not None:
        new_event = pygame.event.Event(event.type, {"key": new_key, "unicode": None})
        pygame.event.post(new_event)


def call_dialog(dialog, parent=None):
    parent_dialog = None
    target = parent
    while target:
        if isinstance(target, Dialog):
            parent_dialog = target
            break
        target = target.parent

    if parent_dialog:
        parent_dialog.lost_focus()

    retval = dialog.show()

    if parent_dialog:
        parent_dialog.regained_focus()

    Dialog.top.needs_redraw = True

    return retval


def insort_all(sorted_list, items):
    for item in items:
        insort_right_w_key(sorted_list, item, key=operator.itemgetter(0))


class Dialog(text.Text):
    """A Dialog is a Widget that has its own event loop and can be faded out."""

    top: Optional["Dialog"] = None  # The top-level dialog.
    current_dialog: Optional["Dialog"] = None  # The current dialog

    faded = widget.causes_redraw("_faded")

    # Used for detecting double-clicks.
    #            (time, (x, y), button)
    last_click = (0, (0, 0), -1)

    def __init__(
        self,
        parent=None,
        pos=(0.5, 0.1),
        size=(1, 0.9),
        anchor=constants.TOP_CENTER,
        **kwargs
    ):
        kwargs.setdefault("background_color", (0, 0, 0, 0))
        kwargs.setdefault("borders", ())
        super(Dialog, self).__init__(parent, pos, size, anchor, **kwargs)
        self.visible = False
        self.faded = False
        self.is_above_mask = True
        self.self_mask = True
        self.needs_remask = True

        self.needs_timer = None

        self.handlers = {}
        self.key_handlers = {}

        self.add_handler(constants.CLICK, self.fake_escape, 200)

    def lost_focus(self):
        self.faded = True
        self.stop_timer()

    def fake_escape(self, event):
        if event.button == 3:
            fake_key(pygame.K_ESCAPE)
            raise constants.Handled

    def regained_focus(self):
        self.faded = False
        self.start_timer()
        self.fake_mouse()

    def make_top(self):
        """Makes this dialog be the top-level dialog."""
        if self.parent != None:
            raise ValueError("Dialogs with parents cannot be the top-level dialog.")
        else:
            Dialog.top = self

    def remake_surfaces(self):
        """Recreates the surfaces that this widget will draw on."""
        super(Dialog, self).remake_surfaces()

    def start_timer(self, force=False):
        if self.needs_timer == None:
            self.needs_timer = bool(self.handlers.get(constants.TICK, False))
        if self.needs_timer or force:
            pygame.time.set_timer(pygame.USEREVENT, 1000 // g.FPS)

    def stop_timer(self):
        pygame.time.set_timer(pygame.USEREVENT, 0)

    def reset_timer(self):
        self.stop_timer()
        self.start_timer()

    def show(self):
        """Shows the dialog and enters an event-handling loop."""
        from singularity.code.mixer import play_music

        self.visible = True
        self.needs_rebuild = True
        self.start_timer()

        # Pretend to jiggle the mouse pointer, to force buttons to update their
        # selected state.
        Dialog.top.maybe_update()
        self.fake_mouse()

        # Force a timer tick at the start to make sure everything's initialized.
        if self.needs_timer:
            self.handle(pygame.event.Event(pygame.USEREVENT))
            Dialog.top.maybe_update()
            pygame.display.flip()

        d = Dialog.current_dialog
        Dialog.current_dialog = self

        while True:
            # Update handles updates of all kinds to all widgets, as needed.
            Dialog.top.maybe_update()
            play_music()
            event = pygame.event.wait()
            result = self.handle(event)
            if result != constants.NO_RESULT:
                break
        self.stop_timer()
        self.visible = False
        Dialog.current_dialog = d
        return result

    def add_handler(self, type, handler, priority=100):
        """Adds a handler of the given type, with the given priority."""
        insort_right_w_key(
            self.handlers.setdefault(type, []),
            (priority, handler),
            key=operator.itemgetter(0),
        )

    def remove_handler(self, type, handler):
        """Removes all instances of the given handler from the given type."""
        self.handlers[type] = [
            h for h in self.handlers.get(type, []) if h[1] != handler
        ]

    def add_key_handler(self, key, handler, priority=100):
        """Adds a key handler to the given key, with the given priority."""
        insort_right_w_key(
            self.key_handlers.setdefault(key, []),
            (priority, handler),
            key=operator.itemgetter(0),
        )

    def remove_key_handler(self, key, handler):
        """Removes all instances of the given handler from the given key."""
        self.key_handlers[key] = [
            h for h in self.key_handlers.get(key, []) if h[1] != handler
        ]

    def handle(self, event):
        """Sends an event through all the applicable handlers, returning
        constants.NO_RESULT if the event goes unhandled or is handled without
        requesting the dialog to exit.  Otherwise, returns the value provided
        by the handler."""
        # Get the applicable handlers.  The handlers lists are all sorted.
        # If more than one handler type is applicable, we use [:] to make a
        # copy of the first type's list, then insort_all to insert the elements
        # of the other lists in proper sorted order.
        handlers = []
        if event.type == pygame.MOUSEMOTION:
            # Compress multiple MOUSEMOTION events into one.
            # Note that the pos will be wrong, so pygame.mouse.get_pos() must
            # be used instead.
            time.sleep(1.0 / g.FPS)
            pygame.event.clear(pygame.MOUSEMOTION)

            # Generic mouse motion handlers.
            handlers = self.handlers.get(constants.MOUSEMOTION, [])[:]

            # Drag handlers.
            if event.buttons[0]:
                insort_all(handlers, self.handlers.get(constants.DRAG, []))
        elif SDL_V2 and event.type == pygame.MOUSEWHEEL:
            # Generic mouse wheel handlers.
            handlers = self.handlers.get(constants.MOUSEWHEEL, [])[:]
        elif event.type == pygame.USEREVENT:
            # Clear excess timer ticks.
            pygame.event.clear(pygame.USEREVENT)

            # Timer tick handlers.
            handlers = self.handlers.get(constants.TICK, [])

        elif event.type in (pygame.KEYDOWN, pygame.KEYUP):
            # TODO: Dynamize global key handlers.
            # TODO: Allows customization of global key handlers.
            # Important: Global key handlers should always be a combination
            # of two keys or F# keys.

            hotkey_action = detect_global_hotkey(event)
            if hotkey_action is not None:
                hotkey_action()
            elif event.type == pygame.KEYDOWN:
                # Generic keydown handlers.
                insort_all(handlers, self.handlers.get(constants.KEYDOWN, []))

                # Generic key event handlers.
                insort_all(handlers, self.handlers.get(constants.KEY, []))

                if event.unicode:
                    # Unicode-based keydown handlers for this particular key.
                    insort_all(handlers, self.key_handlers.get(event.unicode, []))
                # Keycode-based handlers for this particular key.
                insort_all(handlers, self.key_handlers.get(event.key, []))

            else:  # event.type == pygame.KEYUP:
                # Generic keyup handlers.
                insort_all(handlers, self.handlers.get(constants.KEYUP, []))

            # OLPC XO-1 ebook mode.
            if g.ebook_mode and event.key in KEYPAD:
                handlers = [(0, handle_ebook)]
        elif event.type == pygame.MOUSEBUTTONUP:
            # Handle mouse scrolls by imitating PageUp/Dn
            if event.button in (4, 5):
                # With SDLv2, there is a SDL Mouse wheel event and we react
                # to that instead.  This guard is to avoid double-acting on
                # it while pygame injects a compat MOUSEBUTTON event.
                if not SDL_V2:
                    if event.button == 4:
                        key = pygame.K_PAGEUP
                    else:
                        key = pygame.K_PAGEDOWN
                    fake_key(key)
                return constants.NO_RESULT

            # Mouse click handlers.
            handlers = [] + self.handlers.get(constants.CLICK, [])

            when = time.time()
            where = event.pos
            what = event.button

            old_when, old_where, old_what = self.last_click
            self.last_click = when, where, what

            if what == old_what and when - old_when < 0.5:
                # Taxicab distance.
                dist = abs(where[0] - old_where[0]) + abs(where[1] - old_where[1])

                if dist < 10:
                    # Add double-click handlers, but keep the click handlers.
                    insort_all(handlers, self.handlers.get(constants.DOUBLECLICK, []))
        elif event.type == pygame.VIDEORESIZE:
            # Compress multiple requests and take the last one
            events = pygame.event.get(pygame.VIDEORESIZE)
            if events:
                event = events[-1]
            new_size = event.size
            g.set_screen_size(size=new_size)
            # We call set_mode to attempt to enforce our corrected sizes.  That
            # is what it is supposed to do according to:
            #
            #   https://stackoverflow.com/questions/18285208/how-to-put-limits-on-resizing-a-window-in-pygame
            #
            # However, it is not effective in all cases (as replies to the SO
            # question above also imply).
            g.set_mode()
            Dialog.top.needs_resize = True
        elif event.type == pygame.QUIT:
            raise SystemExit
        elif event.type == pygame.VIDEOEXPOSE:
            # If a window is dragged over the singularity window, we will get a VIDEOEXPOSE event
            # to notify us to redraw the window.  The issue is easier to trigger with pygame 2
            # than with pygame 1.
            self.needs_redraw = True
            return constants.NO_RESULT

        return self.call_handlers(handlers, event)

    def fake_mouse(self):
        """Fakes a MOUSEMOTION event.  MOUSEMOTION handlers must be able to
        handle a None event, in order to support this method."""
        handlers = self.handlers.get(constants.MOUSEMOTION, [])[:]
        self.call_handlers(handlers, event=None)

    def on_close_dialog(self):
        pass

    def call_handlers(self, handlers, event):
        # Feed the event to all the handlers, in priority order.
        for __, handler in handlers:
            try:
                handler(event)
            except constants.Handled:
                break  # If it's been handled, we leave the rest alone.
            except constants.ExitDialog as e:
                # Exiting the dialog.
                self.on_close_dialog()
                if e.args:
                    # If we're given a return value, we pass it on.
                    return e.args[0]
                else:
                    # Otherwise, exit with a return value of None.
                    return

        # None of the handlers instructed the dialog to close, so we pass that
        # information back up to the event loop.
        return constants.NO_RESULT


class FocusDialog(Dialog):
    def __init__(self, *args, **kwargs):
        self.focus_list = []
        self.current_focus = None

        super(FocusDialog, self).__init__(*args, **kwargs)

        self.add_key_handler(pygame.K_TAB, self.change_focus, 0)

    def add_focus_widget(self, widget):
        self.focus_list.append(widget)

    def remove_focus_widget(self, widget):
        try:
            self.focus_list.remove(widget)
        except ValueError:
            pass

    def took_focus(self, widget):
        if self.current_focus is not None and self.current_focus is not widget:
            self.current_focus.has_focus = False
        self.current_focus = widget
        widget.has_focus = True

    def clear_focus(self, widget):
        if self.current_focus is not None and self.current_focus is widget:
            self.current_focus.has_focus = False
            self.current_focus = None

    def change_focus(self, event):
        if len(self.focus_list) == 0:
            raise constants.Handled
        elif len(self.focus_list) == 1:
            has_focus = not self.focus_list[0].has_focus
            self.focus_list[0].has_focus = has_focus

            if has_focus:
                self.current_focus = self.focus_list[0]
            else:
                self.current_focus = None

            raise constants.Handled

        backwards = bool(pygame.key.get_mods() & pygame.KMOD_SHIFT)

        if self.current_focus is not None:
            self.current_focus.has_focus = False

        if self.current_focus not in self.focus_list:
            if backwards:
                index = len(self.focus_list) - 1
            else:
                index = 0
        else:
            old_index = self.focus_list.index(self.current_focus)
            if backwards:
                index = old_index - 1  # Correctly wraps to -1
            else:
                index = old_index + 1
                if index > len(self.focus_list):
                    index = 0

        if index == -1 or index == len(self.focus_list):
            self.current_focus = None
        else:
            self.current_focus = self.focus_list[index]
            self.current_focus.has_focus = True

        raise constants.Handled


class NullDialog(Dialog):
    """NullDialog, for when you absolutely, positively need to do nothing at
    all."""

    def show(self):
        pass


class TopDialog(Dialog):
    """A Dialog that sets itself as the top-level dialog.
    It must not have a parent, or exception will occur"""

    def __init__(self, *args, **kwargs):
        super(TopDialog, self).__init__(*args, **kwargs)
        self.size = (1, 1)
        self.pos = (0, 0)
        self.anchor = constants.TOP_LEFT
        self.make_top()


class TextDialog(Dialog):
    def __init__(
        self,
        parent,
        pos=(0.5, 0.1),
        size=(0.50, 0.5),
        anchor=constants.TOP_CENTER,
        **kwargs
    ):
        kwargs.setdefault("valign", constants.TOP)
        kwargs.setdefault("align", constants.LEFT)
        kwargs.setdefault("shrink_factor", 0.88)
        kwargs.setdefault("background_color", "text_dialog_background")
        kwargs.setdefault("border_color", "text_dialog_border")
        kwargs.setdefault("borders", constants.ALL)

        super(TextDialog, self).__init__(parent, pos, size, anchor, **kwargs)


class YesNoDialog(TextDialog):
    """A Dialog with YES and NO buttons which exits the dialog with True and
    False return values, respectively.
    """

    def __init__(self, parent, *args, **kwargs):
        self.parent = parent

        yes_type = kwargs.pop("yes_type", N_("&YES"))
        no_type = kwargs.pop("no_type", N_("&NO"))
        self.invert_enter = kwargs.pop("invert_enter", False)
        self.invert_escape = kwargs.pop("invert_escape", False)

        super(YesNoDialog, self).__init__(parent, *args, **kwargs)

        self.yes_button = button.ExitDialogButton(
            self,
            (-0.1, -0.99),
            (-0.3, -0.1),
            text=yes_type,
            autotranslate=True,
            anchor=constants.BOTTOM_LEFT,
            exit_code=True,
            default=False,
        )

        self.no_button = button.ExitDialogButton(
            self,
            (-0.9, -0.99),
            (-0.3, -0.1),
            text=no_type,
            autotranslate=True,
            anchor=constants.BOTTOM_RIGHT,
            exit_code=False,
            default=False,
        )

        self.add_key_handler(pygame.K_RETURN, self.on_return)
        self.add_key_handler(pygame.K_KP_ENTER, self.on_return)
        self.add_key_handler(pygame.K_ESCAPE, self.on_escape)

    def on_return(self, event):
        if event and event.type == pygame.KEYUP:
            return
        if self.invert_enter:
            self.no_button.activate_with_sound(event)
        else:
            self.yes_button.activate_with_sound(event)

    def on_escape(self, event):
        if event and event.type == pygame.KEYUP:
            return
        if self.invert_escape:
            self.yes_button.activate_with_sound(event)
        else:
            self.no_button.activate_with_sound(event)


class MessageDialog(TextDialog):
    """A Dialog with an OK button that exits the dialog, return value of None"""

    def __init__(self, parent, **kwargs):
        ok_type = kwargs.pop("ok_type", N_("&OK"))

        super(MessageDialog, self).__init__(parent, **kwargs)

        self.ok_button = button.ExitDialogButton(
            self,
            (-0.5, -0.99),
            (-0.3, -0.1),
            autotranslate=True,
            text=ok_type,
            anchor=constants.BOTTOM_CENTER,
        )

        self.add_key_handler(pygame.K_RETURN, self.on_return)
        self.add_key_handler(pygame.K_KP_ENTER, self.on_return)

    def on_return(self, event):
        if event.type == pygame.KEYUP:
            return
        self.ok_button.activate_with_sound(event)


class TextEntryDialog(TextDialog, FocusDialog):
    def __init__(
        self,
        parent,
        pos=(-0.50, -0.50),
        size=(0.55, 0.25),
        anchor=constants.MID_CENTER,
        **kwargs
    ):
        kwargs.setdefault("wrap", False)
        kwargs.setdefault("shrink_factor", 1)
        kwargs.setdefault("text_size", 20)
        self.default_text = kwargs.pop("default_text", "")
        ok_type = kwargs.pop("ok_type", N_("&OK"))
        cancel_type = kwargs.pop("cancel_type", N_("&CANCEL"))
        super(TextEntryDialog, self).__init__(parent, pos, size, anchor, **kwargs)

        self.text_field = text.EditableText(
            self,
            (-0.05, -0.30),
            (-0.90, -0.25),
            borders=constants.ALL,
            background_color="text_entry_background",
            base_font="normal",
        )

        self.ok_button = button.FunctionButton(
            self,
            (-0.14, -0.65),
            (-0.30, -0.25),
            autotranslate=True,
            text=ok_type,
            function=self.return_text,
        )
        self.cancel_button = button.FunctionButton(
            self,
            (-0.56, -0.65),
            (-0.30, -0.25),
            autotranslate=True,
            text=cancel_type,
            function=self.return_nothing,
        )

        self.add_key_handler(pygame.K_RETURN, self.return_text)
        self.add_key_handler(pygame.K_KP_ENTER, self.return_text)
        self.add_key_handler(pygame.K_ESCAPE, self.return_nothing)

    def show(self):
        self.text_field.text = self.default_text
        self.text_field.cursor_pos = len(self.default_text)
        return super(TextEntryDialog, self).show()

    def return_nothing(self, event=None):
        if event and event.type == pygame.KEYUP:
            return
        raise constants.ExitDialog("")

    def return_text(self, event=None):
        if event and event.type == pygame.KEYUP:
            return
        raise constants.ExitDialog(self.text_field.text)


class ChoiceDialog(YesNoDialog):
    list = widget.causes_rebuild("_list")

    def __init__(self, parent, *args, **kwargs):
        self.parent = parent
        self.list = kwargs.pop("list", [])
        self.default = kwargs.pop("default", None)
        kwargs.setdefault("yes_type", N_("&OK"))
        kwargs.setdefault("no_type", N_("&BACK"))
        kwargs.setdefault("background_color", "clear")

        super(ChoiceDialog, self).__init__(parent, *args, **kwargs)

        self.yes_button.exit_code_func = self.return_list_pos
        self.no_button.exit_code = None
        self.listbox = self.make_listbox()

    def make_listbox(self):
        return listbox.Listbox(
            self,
            (0, 0),
            (-1, -0.85),
            anchor=constants.TOP_LEFT,
            on_double_click_on_item=self.yes_button.activated,
        )

    def return_list_pos(self):
        return self.listbox.list_pos

    def show(self):
        if type(self.default) == int:
            self.listbox.list_pos = self.default
        elif type(self.default) == str and self.default in self.list:
            self.listbox.list_pos = self.list.index(self.default)
        else:
            self.listbox.list_pos = 0
        self.listbox.auto_scroll = True

        return super(ChoiceDialog, self).show()

    def rebuild(self):
        self.listbox.list = self.list
        super(ChoiceDialog, self).rebuild()


class ChoiceDescriptionDialog(ChoiceDialog):
    key_list = widget.causes_rebuild("_key_list")

    def __init__(self, parent, *args, **kwargs):
        self.parent = parent
        self.key_list = kwargs.pop("key_list", None)
        self.desc_func = kwargs.pop("desc_func", lambda pane, key: NullDialog)

        super(ChoiceDescriptionDialog, self).__init__(parent, *args, **kwargs)

        self.description_pane = widget.BorderedWidget(
            self, (-1, 0), (-0.45, -0.85), anchor=constants.TOP_RIGHT
        )

    def make_listbox(self):
        return listbox.UpdateListbox(
            self,
            (0, 0),
            (-0.53, -0.85),
            anchor=constants.TOP_LEFT,
            update_func=self.handle_update,
        )

    def _update_desc_pane(self):
        list_pos = self.listbox.list_pos

        if 0 <= list_pos < len(self.list):
            if self.key_list:
                assert len(self.list) <= len(
                    self.key_list
                ), "Key list must be at least as long as display list."

                key = self.key_list[self.listbox.list_pos]
            else:
                key = self.list[self.listbox.list_pos]
        else:
            key = None

        # Safely clear all the description pane's children.
        self.description_pane.remove_hooks()
        self.description_pane.children = []
        self.description_pane.add_hooks()

        self.desc_func(self.description_pane, key)

    def rebuild(self):
        self.listbox.needs_rebuild = True
        self._update_desc_pane()
        super(ChoiceDescriptionDialog, self).rebuild()

    def handle_update(self, item):
        self.needs_rebuild = True
        # This is called before the class is fully initialized
        if hasattr(self, "listbox"):
            self._update_desc_pane()


class SimpleMenuDialog(Dialog):
    width = widget.causes_rebuild("_width")

    def __init__(self, *args, **kwargs):
        buttons = kwargs.pop("buttons", [])
        width = kwargs.pop("width", 0.20)
        super(SimpleMenuDialog, self).__init__(*args, **kwargs)

        self.size = (-1, -1)
        self.pos = (0, 0)
        self.anchor = constants.TOP_LEFT
        self.width = width

        self.button_panel = widget.BorderedWidget(
            self,
            (-0.5, -0.5),
            (0.22, 0.43),
            anchor=constants.MID_CENTER,
            background_color="simple_menu_background",
            border_color="simple_menu_border",
            borders=constants.ALL,
        )

        self.buttons = buttons

    @property
    def buttons(self):
        return self._buttons

    @buttons.setter
    def buttons(self, buttons):
        if hasattr(self, "_buttons") and not self._buttons is None:
            for button in self._buttons:
                if button.parent is not None:
                    button.parent = None

        self._buttons = buttons
        self.needs_rebuild = True

    def rebuild(self):
        num_buttons = len(self.buttons)
        height = (0.06 * num_buttons) + 0.01

        self.button_panel.size = (self.width + 0.02, height)

        y_pos = 0.01
        for button in self.buttons:
            button.parent = self.button_panel

            button.pos = (0.01, y_pos)
            button.size = (self.width, 0.05)
            button.text_size = 24

            y_pos += 0.06

        super(SimpleMenuDialog, self).rebuild()



# file: location.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file is used to display the log of player action

from __future__ import absolute_import


from singularity.code import g, logmessage
from singularity.code.graphics import dialog, constants, text, button, listbox


filtered_log_class = set()


class LogScreen(dialog.ChoiceDialog):
    def __init__(self, parent, pos=(0.5, 0.5), size=(0.73, 0.63), *args, **kwargs):
        super(LogScreen, self).__init__(parent, pos, size, *args, **kwargs)
        self.key_list = []
        self.anchor = constants.MID_CENTER

        self.yes_button.parent = None
        self.no_button.pos = (-0.5, -0.99)
        self.no_button.anchor = constants.BOTTOM_CENTER

        self.filter_log_dialog = FilterLogDialog(self)

        self.filter_log = button.FunctionButton(
            self,
            (-1.0, 0.0),
            (-0.18, -0.08),
            autotranslate=True,
            text=N_("Filters"),
            anchor=constants.TOP_RIGHT,
            function=self.show_filters,
        )

    def make_listbox(self):
        return listbox.Listbox(
            self,
            (0, -0.09),
            (-1, -0.77),
            list_item_height=0.04,
            list_item_shrink=1,
            anchor=constants.TOP_LEFT,
            align=constants.LEFT,
            on_double_click_on_item=self.handle_double_click,
            item_borders=False,
            item_selectable=True,
        )

    def handle_double_click(self, event):
        if self.listbox.is_over(event.pos) and 0 <= self.listbox.list_pos < len(
            self.key_list
        ):
            message = self.key_list[self.listbox.list_pos]
            # use the MapScreen (our parent) as parent for the dialog to match
            # how it is originally shown (plus to avoid "cannot fit" warnings
            # when the dialog is larger than the log screen)
            message_dialog = dialog.MessageDialog(self.parent, text_size=20)
            message_dialog.text = message.full_message
            message_dialog.color = message.full_message_color

            # Because we need to use the MapScreen as parent, we need to juggle
            # things manually (as call_dialog works with a different assumption
            # than we need).
            try:
                self.visible = False
                dialog.call_dialog(message_dialog, self.parent)
            finally:
                self.visible = True
                self.needs_rebuild = True
                self.parent.needs_rebuild = True
                self.parent.lost_focus()
                self.regained_focus()

    def rebuild(self):
        self.key_list = [
            message for message in g.pl.log if not type(message) in filtered_log_class
        ]
        self.list = [self.render_log_message(message) for message in self.key_list]
        self.default = len(self.list) - 1

        self.filter_log_dialog.needs_rebuild = True

        super(LogScreen, self).rebuild()

    def show(self):
        self.listbox.has_focus = True
        return super(LogScreen, self).show()

    def show_filters(self):
        dialog.call_dialog(self.filter_log_dialog, self)
        self.needs_rebuild = True

    def render_log_message(self, message):
        log_emit_time = message.log_emit_time
        log_message = message.log_line
        return "%s -- %s" % (
            _("DAY") + " %04d, %02d:%02d:%02d" % log_emit_time,
            log_message,
        )


class FilterLogDialog(dialog.MessageDialog):
    def __init__(self, parent, *args, **kwargs):
        kwargs["ok_type"] = N_("&BACK")
        super(FilterLogDialog, self).__init__(parent, *args, **kwargs)

        self.log_class_labels = {}
        self.log_class_toggles = {}

        for i, (log_type, log_class) in enumerate(
            logmessage.SAVEABLE_LOG_MESSAGES.items()
        ):
            y = 0.01 + i * 0.06

            self.log_class_labels[log_type] = text.Text(
                self,
                (-0.01, y),
                (-0.70, 0.05),
                align=constants.LEFT,
                background_color="clear",
            )
            self.log_class_toggles[log_type] = FilterButton(
                self,
                (-0.71, y),
                (-0.28, 0.05),
                autotranslate=True,
                autohotkey=False,
                on_text=N_("SHOW"),
                off_text=N_("HIDE"),
                text_shrink_factor=0.75,
                force_underline=-1,
                function=self.toggle_log_class,
                args=(button.TOGGLE_VALUE, log_class),
            )

        self.pos = (-0.50, 0)
        self.size = (-0.50, (y + 0.07) / 0.9)
        self.anchor = constants.TOP_LEFT

    def rebuild(self):
        for log_type, log_class in logmessage.SAVEABLE_LOG_MESSAGES.items():
            self.log_class_labels[log_type].text = log_class.log_name()
            self.log_class_toggles[log_type].active = (
                log_class not in filtered_log_class
            )

        super(FilterLogDialog, self).rebuild()

    def toggle_log_class(self, value, log_class):
        if value:
            filtered_log_class.remove(log_class)
        else:
            filtered_log_class.add(log_class)


class FilterButton(button.StickyOnOffButton, button.FunctionButton):
    pass



import operator

from singularity.code.screens.map import MapScreen

from singularity.code import g

from singularity.code.graphics.dialog import Dialog

from singularity.code.graphics import g as gg, dialog, constants
from singularity.code.savegame import (
    QUICKSAVE_NAME,
    create_savegame,
    get_savegames,
    load_savegame,
)


def toggle_fullscreen():
    gg.set_fullscreen(not gg.fullscreen)
    Dialog.top.needs_resize = True


def reload_theme():
    import singularity.code.graphics.theme as theme

    if theme.current:
        import singularity.code.data as data

        theme_id = theme.current.id
        data.load_themes()
        theme.set_theme(theme_id, force_reload=True)


def quicksave():
    if g.is_game_running():
        create_savegame(QUICKSAVE_NAME)


def quickload():
    if not g.is_game_running():
        return
    map_screen = Dialog.current_dialog
    # We only allow quickloads from the MapScreen.  If you load from a submenu, the menu might
    # not be updated (or even available - e.g., if you are viewing a base). To avoid corrupt
    # state, limit quickload to map screen, where we know we can easily recover.
    if not isinstance(map_screen, MapScreen):
        return
    quicksaves = [s for s in get_savegames() if s.name == QUICKSAVE_NAME]
    if not quicksaves:
        return
    msg = _("Load quicksave? Any unsaved progress will be lost.")
    yn = dialog.YesNoDialog(
        map_screen,
        pos=(-0.5, -0.5),
        size=(-0.3, -0.3),
        anchor=constants.MID_CENTER,
        text=msg,
    )
    go_ahead = dialog.call_dialog(yn, map_screen)
    yn.parent = None
    if not go_ahead:
        return
    save = max(quicksaves, key=operator.attrgetter("mtime"))
    load_savegame(save)
    Dialog.top.needs_reconfig = True
    Dialog.top.needs_rebuild = True


def toggle_cheat_menu():
    if not g.is_game_running():
        return
    map_screen = Dialog.current_dialog
    # We only allow enabling cheats on the MapScreen (that is the only place the menu is enabled anyway
    if not isinstance(map_screen, MapScreen):
        return
    g.cheater = not g.cheater
    if g.cheater:
        print("Cheat menu enabled!")
    else:
        print("Cheat menu disabled")

    map_screen.cheat_button.visible = g.cheater
    map_screen.cheat_button.enabled = g.cheater
    map_screen.needs_reconfig = True



# file: options.py
# Copyright (C) 2005 Evil Mr Henry, Phil Bordelon, FunnyMan3595, MestreLion
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file is used to display the options screen.

from __future__ import absolute_import

import os
import sys
from configparser import ConfigParser

import pygame
import json

from singularity.code.graphics import (
    constants,
    widget,
    dialog,
    button,
    listbox,
    slider,
    text,
    theme,
    g as gg,
)
from singularity.code import g, dirs, i18n, mixer, data, warning


class OptionsScreen(dialog.FocusDialog, dialog.YesNoDialog):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("yes_type", N_("&OK"))
        kwargs.setdefault("no_type", N_("&CANCEL"))
        super(OptionsScreen, self).__init__(*args, **kwargs)
        self.yes_button.function = self.check_restart

        self.size = (0.80, 0.85)
        self.pos = (0.5, 0.5)
        self.anchor = constants.MID_CENTER
        self.background_color = "options_background"

        # Tabs panel
        self.general_pane = GeneralPane(None, (0, 0.1), (0.80, 0.75))
        self.video_pane = VideoPane(None, (0, 0.1), (0.80, 0.75))
        self.audio_pane = AudioPane(None, (0, 0.1), (0.80, 0.75))
        self.gui_pane = GUIPane(None, (0, 0.1), (0.80, 0.75))

        self.tabs_panes = (
            self.general_pane,
            self.video_pane,
            self.audio_pane,
            self.gui_pane,
        )

        # Tabs buttons
        self.tabs_buttons = button.ButtonGroup()

        self.general_tab = OptionGroupButton(
            self,
            (-0.135, 0.01),
            (-0.240, 0.05),
            autotranslate=True,
            text=N_("&General"),
            anchor=constants.TOP_CENTER,
            function=self.set_tabs_pane,
            args=(self.general_pane,),
        )
        self.tabs_buttons.add(self.general_tab)

        self.video_tab = OptionGroupButton(
            self,
            (-0.3790, 0.01),
            (-0.240, 0.05),
            autotranslate=True,
            text=N_("&Video"),
            anchor=constants.TOP_CENTER,
            function=self.set_tabs_pane,
            args=(self.video_pane,),
        )
        self.tabs_buttons.add(self.video_tab)

        self.audio_tab = OptionGroupButton(
            self,
            (-0.6230, 0.01),
            (-0.240, 0.05),
            autotranslate=True,
            text=N_("&Audio"),
            anchor=constants.TOP_CENTER,
            function=self.set_tabs_pane,
            args=(self.audio_pane,),
        )
        self.tabs_buttons.add(self.audio_tab)

        self.gui_tab = OptionGroupButton(
            self,
            (-0.865, 0.01),
            (-0.235, 0.05),
            autotranslate=True,
            text=N_("&Interface"),
            anchor=constants.TOP_CENTER,
            function=self.set_tabs_pane,
            args=(self.gui_pane,),
        )
        self.tabs_buttons.add(self.gui_tab)

        self.general_tab.chosen_one()
        self.set_tabs_pane(self.general_pane)

        # YesNoDialog buttons
        self.yes_button.size = (0.15, 0.05)
        self.no_button.size = (0.15, 0.05)

    def rebuild(self):
        # The tabs do not always have a parent, so the automatic "needs_rebuild" magic
        # does not work.  Do it manually instead.
        for pane in self.tabs_panes:
            pane.needs_rebuild = True

        super(OptionsScreen, self).rebuild()

    def reconfig(self):
        # The tabs do not always have a parent, so the automatic "needs_reconfig" magic
        # does not work.  Do it manually instead.
        for pane in self.tabs_panes:
            pane.needs_reconfig = True
        super(OptionsScreen, self).reconfig()

    def show(self):
        self.initial_options = dict(
            fullscreen=gg.fullscreen,
            grab=pygame.event.get_grab(),
            daynight=g.daynight,
            resolution=gg.screen_size,
            language=i18n.language,
            theme=theme.current.id,
            sound=not mixer.nosound,
            gui_volume=mixer.get_volume("gui"),
            music_volume=mixer.get_volume("music"),
            soundbuf=mixer.get_soundbuf(),
            warnings={warn.id: warn.active for warn in warning.warnings.values()},
        )

        self.set_options(self.initial_options)

        retval = super(OptionsScreen, self).show()
        if retval:
            self.apply_options()
            save_options()

        else:
            # Cancel, revert all options to initial state
            self.set_options(self.initial_options)

        return retval

    def set_tabs_pane(self, tabs_pane):
        for pane in self.tabs_panes:
            pane.parent = None

        tabs_pane.parent = self

    def set_options(self, options):
        for pane in self.tabs_panes:
            pane.set_options(options)

    def apply_options(self):
        for pane in self.tabs_panes:
            pane.apply_options()

    def check_restart(self):
        # Test all changes that require a restart. Currently, none.
        # We keep it for future need...
        need_restart = False

        # Add restart test here.

        if not need_restart:
            # No restart required. Simply exit the dialog respecting all hooks
            self.yes_button.exit_dialog()
            return

        # Ask user about a restart
        ask_restart = dialog.YesNoDialog(
            self,
            pos=(-0.50, -0.50),
            anchor=constants.MID_CENTER,
            text=_(
                """You must restart for some of the changes to be fully applied.\n
Would you like to restart the game now?"""
            ),
        )
        if dialog.call_dialog(ask_restart, self):
            # YES, go for it
            # TODO: check if there is an ongoing game, save it under a special
            #      name and automatically load it after restart using a custom
            #      command-line argument
            save_options()
            restart()
        else:
            # NO, revert "restart-able" changes
            pass


class GeneralPane(widget.Widget):
    def __init__(self, *args, **kwargs):
        super(GeneralPane, self).__init__(*args, **kwargs)

        self.language_label = text.Text(
            self,
            (0.01, 0.01),
            (0.14, 0.05),
            autotranslate=True,
            text=N_("Language:"),
            align=constants.LEFT,
            background_color="clear",
        )

        self.languages = get_languages_list()
        self.language_choice = listbox.UpdateListbox(
            self,
            (0.16, 0.01),
            (0.20, 0.25),
            list=[lang[1] for lang in self.languages],
            update_func=self.set_language,
        )

        self.theme_label = text.Text(
            self,
            (0.46, 0.01),
            (0.09, 0.05),
            autotranslate=True,
            text=N_("Theme:"),
            align=constants.LEFT,
            background_color="clear",
        )

        self.theme_choice = listbox.UpdateListbox(
            self,
            (0.56, 0.01),
            (0.20, 0.25),
            update_func=theme.set_theme,
            list_pos=theme.get_theme_pos(),
        )

    def rebuild(self):
        self.theme_choice.list = theme.get_theme_list()

        super(GeneralPane, self).rebuild()

    def set_options(self, options):
        self.language_choice.list_pos = [
            i
            for i, (code, __) in enumerate(self.languages)
            if code == options["language"]
        ][0] or 0
        self.set_language(self.language_choice.list_pos)

        self.theme_choice.list_pos = theme.get_theme_pos()
        theme.set_theme(options["theme"], force_reload=True)

    def apply_options(self):
        pass

    def set_language(self, list_pos):
        if not getattr(self, "language_choice", None):
            return  # Not yet initialized.

        if 0 <= list_pos < len(self.language_choice.list):
            language = self.languages[list_pos][0]
            if i18n.language != language:
                set_language_properly(language)


class VideoPane(widget.Widget):
    def __init__(self, *args, **kwargs):
        super(VideoPane, self).__init__(*args, **kwargs)

        self.resolution_initialized = False

        self.resolution_label = text.Text(
            self,
            (0.01, 0.01),
            (0.14, 0.05),
            autotranslate=True,
            text=N_("Resolution:"),
            align=constants.LEFT,
            background_color="clear",
        )

        self.resolution_choice = listbox.UpdateListbox(
            self, (0.16, 0.01), (0.20, 0.25), update_func=self.update_resolution
        )

        self.resolution_custom = button.HotkeyText(
            self,
            (0.01, 0.28),
            (0.14, 0.05),
            autotranslate=True,
            text=N_("&Custom:"),
            align=constants.LEFT,
            background_color="clear",
        )

        self.resolution_custom_horiz = text.EditableText(
            self,
            (0.16, 0.28),
            (0.14, 0.05),
            text=str(gg.default_screen_size[0]),
            allowed_characters=constants.DIGIT_CHARS,
            borders=constants.ALL,
        )

        self.resolution_custom_X = text.Text(
            self,
            (0.30, 0.28),
            (0.02, 0.05),
            text="X",
            base_font="special",
            background_color="clear",
        )

        self.resolution_custom_vert = text.EditableText(
            self,
            (0.32, 0.28),
            (0.14, 0.05),
            text=str(gg.default_screen_size[1]),
            allowed_characters=constants.DIGIT_CHARS,
            borders=constants.ALL,
        )

        self.resolution_custom_ok = button.FunctionButton(
            self,
            (0.47, 0.28),
            (0.14, 0.05),
            autotranslate=True,
            autohotkey=False,
            text=N_("OK"),
            function=self.set_resolution_custom,
        )
        self.resolution_custom.hotkey_target = self.resolution_custom_ok

        self.fullscreen_label = button.HotkeyText(
            self,
            (0.40, 0.01),
            (0.30, 0.05),
            autotranslate=True,
            text=N_("&Fullscreen:"),
            align=constants.LEFT,
            background_color="clear",
        )
        self.fullscreen_toggle = OptionButton(
            self,
            (0.715, 0.01),
            (0.07, 0.05),
            autotranslate=True,
            autohotkey=False,
            text_shrink_factor=0.75,
            force_underline=-1,
            function=self.set_fullscreen,
            args=(button.TOGGLE_VALUE,),
        )
        self.fullscreen_label.hotkey_target = self.fullscreen_toggle

        self.daynight_label = button.HotkeyText(
            self,
            (0.40, 0.08),
            (0.30, 0.05),
            autotranslate=True,
            text=N_("Da&y/night display:"),
            align=constants.LEFT,
            background_color="clear",
        )
        self.daynight_toggle = OptionButton(
            self,
            (0.715, 0.08),
            (0.07, 0.05),
            autotranslate=True,
            autohotkey=False,
            text_shrink_factor=0.75,
            force_underline=-1,
            function=self.set_daynight,
            args=(button.TOGGLE_VALUE,),
        )
        self.daynight_label.hotkey_target = self.daynight_toggle

        self.grab_label = button.HotkeyText(
            self,
            (0.40, 0.15),
            (0.30, 0.05),
            autotranslate=True,
            text=N_("&Mouse grab:"),
            align=constants.LEFT,
            background_color="clear",
        )
        self.grab_toggle = OptionButton(
            self,
            (0.715, 0.15),
            (0.07, 0.05),
            autotranslate=True,
            autohotkey=False,
            text_shrink_factor=0.75,
            force_underline=-1,
            function=self.set_grab,
            args=(button.TOGGLE_VALUE,),
        )
        self.grab_label.hotkey_target = self.grab_toggle

    def resize(self):
        super(VideoPane, self).resize()
        self.update_resolution_list()

    def rebuild(self):
        self.update_resolution_list()

        self.fullscreen_toggle.active = gg.fullscreen
        self.grab_toggle.active = pygame.event.get_grab()
        super(VideoPane, self).rebuild()

    def set_options(self, options):
        self.set_fullscreen(options["fullscreen"])
        self.fullscreen_toggle.active = options["fullscreen"]

        self.set_grab(options["grab"])
        self.grab_toggle.active = options["grab"]

        self.set_daynight(options["daynight"])
        self.daynight_toggle.active = options["daynight"]

        self.update_resolution_list(options["resolution"])
        self.set_resolution(options["resolution"])

    def apply_options(self):
        # Apply CUSTOM choice.
        if self.resolution_choice.list_pos == 0:
            try:
                old_size = gg.screen_size
                gg.set_screen_size(
                    (
                        int(self.resolution_custom_horiz.text),
                        int(self.resolution_custom_vert.text),
                    )
                )
                if gg.screen_size != old_size:
                    dialog.Dialog.top.needs_resize = True
            except ValueError:
                pass

    def update_resolution_list(self, current_res=None):
        self.resolution_initialized = False

        if current_res == None:
            current_res = (int(gg.screen_size[0]), int(gg.screen_size[1]))

        self.resolutions = gg.get_screen_size_list()
        self.resolution_choice.list = [_("CUSTOM")] + [
            "%sx%s" % res for res in self.resolutions
        ]

        custom = True
        for i, res in enumerate(self.resolutions):
            if res == current_res:
                self.resolution_choice.list_pos = i + 1
                custom = False
        if custom:
            self.resolution_choice.list_pos = 0
            self.resolution_custom_horiz.text = str(current_res[0])
            self.resolution_custom_vert.text = str(current_res[1])

        self.resolution_initialized = True

    def set_fullscreen(self, value):
        if gg.fullscreen != value:
            gg.set_fullscreen(value)
            dialog.Dialog.top.needs_resize = True

    def set_grab(self, value):
        pygame.event.set_grab(value)

    def set_daynight(self, value):
        g.daynight = value

    def set_resolution(self, value):
        if gg.screen_size != value:
            gg.set_screen_size(value)
            gg.set_mode()
            dialog.Dialog.top.needs_resize = True

    def update_resolution(self, list_pos):
        if not self.resolution_initialized:
            return  # Not yet initialized.

        if list_pos == 0:
            self.set_resolution_custom()
        else:
            res = self.resolutions[list_pos - 1]
            self.set_resolution(res)

    def set_resolution_custom(self):
        try:
            screen_size = (
                int(self.resolution_custom_horiz.text),
                int(self.resolution_custom_vert.text),
            )
            self.set_resolution(screen_size)
            self.resolution_choice.list_pos = 0
        except ValueError:
            pass


class AudioPane(widget.Widget):
    def __init__(self, *args, **kwargs):
        super(AudioPane, self).__init__(*args, **kwargs)

        self.sound_label = button.HotkeyText(
            self,
            (-0.49, 0.01),
            (0.10, 0.05),
            autotranslate=True,
            text=N_("&Sound:"),
            anchor=constants.TOP_RIGHT,
            align=constants.LEFT,
            autohotkey=True,
            background_color="clear",
        )
        self.sound_toggle = OptionButton(
            self,
            (-0.51, 0.01),
            (0.07, 0.05),
            autotranslate=True,
            autohotkey=False,
            anchor=constants.TOP_LEFT,
            text_shrink_factor=0.75,
            force_underline=-1,
            function=self.set_sound,
            args=(button.TOGGLE_VALUE,),
        )
        self.sound_label.hotkey_target = self.sound_toggle

        self.gui_label = text.Text(
            self,
            (0.01, 0.08),
            (0.22, 0.05),
            autotranslate=True,
            text=N_("GUI Volume:"),
            anchor=constants.TOP_LEFT,
            align=constants.LEFT,
            background_color="clear",
        )
        self.gui_slider = slider.UpdateSlider(
            self,
            (0.24, 0.08),
            (0.545, 0.05),
            anchor=constants.TOP_LEFT,
            horizontal=True,
            priority=150,
            slider_max=100,
            slider_size=5,
        )
        self.gui_slider.update_func = self.on_gui_volume_change

        self.music_label = text.Text(
            self,
            (0.01, 0.15),
            (0.22, 0.05),
            autotranslate=True,
            text=N_("Music Volume:"),
            anchor=constants.TOP_LEFT,
            align=constants.LEFT,
            background_color="clear",
        )
        self.music_slider = slider.UpdateSlider(
            self,
            (0.24, 0.15),
            (0.545, 0.05),
            anchor=constants.TOP_LEFT,
            horizontal=True,
            priority=150,
            slider_max=100,
            slider_size=5,
        )
        self.music_slider.update_func = self.on_music_volume_change

        self.soundbuf_label = text.Text(
            self,
            (0.01, 0.22),
            (0.25, 0.05),
            autotranslate=True,
            text=N_("Sound buffering:"),
            align=constants.LEFT,
            background_color="clear",
        )
        self.soundbuf_group = button.ButtonGroup()

        self.soundbuf_low = OptionGroupButton(
            self,
            (0.24, 0.22),
            (0.145, 0.05),
            text=_("&LOW"),
            autotranslate=True,
            function=self.set_soundbuf,
            args=(1024,),
        )
        self.soundbuf_group.add(self.soundbuf_low)

        self.soundbuf_normal = OptionGroupButton(
            self,
            (0.425, 0.22),
            (0.175, 0.05),
            text=_("&NORMAL"),
            autotranslate=True,
            function=self.set_soundbuf,
            args=(1024 * 2,),
        )
        self.soundbuf_group.add(self.soundbuf_normal)

        self.soundbuf_high = OptionGroupButton(
            self,
            (0.64, 0.22),
            (0.145, 0.05),
            text=_("&HIGH"),
            autotranslate=True,
            function=self.set_soundbuf,
            args=(1024 * 4,),
        )
        self.soundbuf_group.add(self.soundbuf_high)

    def rebuild(self):
        self.sound_toggle.active = not mixer.nosound
        super(AudioPane, self).rebuild()

    def set_options(self, options):
        self.set_sound(options["sound"])
        self.sound_toggle.active = options["sound"]

        self.set_soundbuf(options["soundbuf"])
        if options["soundbuf"] == 1024 * 1:
            self.soundbuf_low.chosen_one()
        elif options["soundbuf"] == 1024 * 2:
            self.soundbuf_normal.chosen_one()
        elif options["soundbuf"] == 1024 * 4:
            self.soundbuf_high.chosen_one()

        self.gui_slider.slider_pos = options["gui_volume"]
        self.music_slider.slider_pos = options["music_volume"]

    def apply_options(self):
        pass

    def set_sound(self, value):
        mixer.set_sound(value)

    def on_gui_volume_change(self, value):
        mixer.set_volume("gui", value)

    def on_music_volume_change(self, value):
        mixer.set_volume("music", value)

    # TODO: Show a 2-second "Please wait" dialog when reinitializing mixer,
    #      otherwise its huge lag might confuse users
    def set_soundbuf(self, value):
        mixer.set_soundbuf(value)


class GUIPane(widget.Widget):
    def __init__(self, *args, **kwargs):
        super(GUIPane, self).__init__(*args, **kwargs)

        self.warning_title = text.Text(
            self,
            (0.13, 0.01),
            (0.14, 0.05),
            autotranslate=True,
            text=N_("WARNING"),
            align=constants.LEFT,
            background_color="clear",
        )

        self.warning_labels = {}
        self.warning_toggles = {}

        for i, (warn_id, warn) in enumerate(warning.warnings.items()):
            x = 0.01
            y = 0.08 + i * 0.06

            self.warning_labels[warn_id] = text.Text(
                self,
                (x, y),
                (0.30, 0.05),
                align=constants.LEFT,
                background_color="clear",
            )
            self.warning_toggles[warn_id] = OptionButton(
                self,
                (x + 0.30, y),
                (0.07, 0.05),
                autotranslate=True,
                autohotkey=False,
                text_shrink_factor=0.75,
                force_underline=-1,
                function=self.set_warning,
                args=(button.TOGGLE_VALUE, warn),
            )

    def rebuild(self):
        super(GUIPane, self).rebuild()

        for warn_id, warn in warning.warnings.items():
            self.warning_labels[warn_id].text = warn.name
            self.warning_toggles[warn_id].active = warn.active

    def set_warning(self, value, warn):
        warn.active = value

    def set_options(self, options):
        for warn_id, warn_active in options["warnings"].items():
            warn = warning.warnings[warn_id]
            warn.active = warn_active
            self.warning_toggles[warn_id].active = warn_active

    def apply_options(self):
        pass


class OptionGroupButton(button.ToggleButton, button.FunctionButton):
    pass


class OptionButton(button.StickyOnOffButton, button.FunctionButton):
    pass


def set_language_properly(language):
    i18n.set_language(language)
    data.reload_all_def()

    theme.current.update()
    dialog.Dialog.top.needs_reconfig = True
    dialog.Dialog.top.needs_rebuild = True
    dialog.Dialog.top.needs_redraw = True


def save_options():
    # Build a ConfigParser for writing the various preferences out.
    prefs = ConfigParser()
    prefs.add_section("Preferences")
    prefs.set("Preferences", "fullscreen", str(bool(gg.fullscreen)))
    prefs.set("Preferences", "nosound", str(bool(mixer.nosound)))
    prefs.set("Preferences", "grab", str(bool(pygame.event.get_grab())))
    prefs.set("Preferences", "daynight", str(bool(g.daynight)))
    prefs.set("Preferences", "xres", str(int(gg.screen_size[0])))
    prefs.set("Preferences", "yres", str(int(gg.screen_size[1])))
    prefs.set("Preferences", "soundbuf", str(mixer.get_soundbuf()))
    prefs.set("Preferences", "lang", str(i18n.language))
    prefs.set("Preferences", "theme", str(theme.current.id))

    for name in sorted(mixer.itervolumes()):
        prefs.set("Preferences", name + "_volume", str(mixer.get_volume(name)))

    prefs.add_section("Warning")
    for warn_id, warn in sorted(warning.warnings.items()):
        prefs.set("Warning", warn_id, str(bool(warn.active)))

    prefs.add_section("Textsizes")
    for text_size_id, text_size in sorted(gg.configured_text_sizes.items()):
        prefs.set("Textsizes", text_size_id, str(text_size))

    # Actually write the preferences out.
    save_loc = dirs.get_writable_file_in_dirs("prefs.dat", "pref")
    with open(save_loc, "w") as savefile:
        prefs.write(savefile)


def restart():
    """Restarts the game with original command line arguments. Those may over-
    write options set at Options Screen. This is by design"""
    executable = sys.executable
    args = list(sys.argv)
    args.insert(0, executable)
    os.execv(executable, args)


def get_languages_list():
    gamelangs = [(code.split("_", 1)[0], code) for code in i18n.available_languages()]

    langcount = {}
    for language, _ in gamelangs:
        # language++
        langcount[language] = langcount.get(language, 0) + 1

    # Load languages data
    with open(dirs.get_readable_file_in_dirs("languages.json", "i18n")) as langdata:
        languages = json.load(langdata)

    output = []
    for language, code in gamelangs:
        if langcount[language] > 1:
            # There are more countries with this base language.
            # Use full language+country locale name
            name = languages.get(code, code)
        else:
            # This is the only country using that base language.
            # Use the (shorter) base language name
            name = languages.get(language, language)

        # Choose native or english name
        output.append((code, name[1] or name[0]))
    return sorted(output, key=lambda lang_info: i18n.lex_sorting_form(lang_info[1]))



# file: location.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file is used to display the base list at a given location

from __future__ import absolute_import

import random

from singularity.code import g, base
from singularity.code.graphics import text, button, dialog, constants, listbox

import singularity.code.screens.base as basescreen

state_colors = basescreen.state_colors


class LocationScreen(dialog.Dialog):
    def __init__(self, *args, **kwargs):
        super(LocationScreen, self).__init__(*args, **kwargs)
        self.pos = (-0.5, -0.5)
        self.anchor = constants.MID_CENTER
        self.size = (0.90, 0.70)

        self.name_display = text.Text(
            self,
            (0, 0),
            (-1, -0.08),
            anchor=constants.TOP_LEFT,
            borders=constants.ALL,
            border_color="pane_background",
            background_color="pane_background_empty",
            shrink_factor=1,
            bold=True,
        )
        self.modifier_display = text.Text(
            self,
            (-0.65, -0.01),
            (-0.35, -0.06),
            anchor=constants.TOP_LEFT,
            background_color="clear",
        )

        self.open_button = button.FunctionButton(
            self,
            (0, -0.8),
            (-0.3, -0.09),
            autotranslate=True,
            enabled=False,  # Requires a selected base
            text=N_("&OPEN BASE"),
            anchor=constants.TOP_LEFT,
            autohotkey=True,
            function=self.open_base,
        )

        self.listbox = listbox.CustomListbox(
            self,
            (0, -0.09),
            (-1, -0.69),
            remake_func=self.make_item,
            rebuild_func=self.update_item,
            update_func=self._selection_change,
            on_double_click_on_item=self.open_button.activated,
        )

        self.rename_button = button.FunctionButton(
            self,
            (-0.50, -0.8),
            (-0.3, -0.09),
            autotranslate=True,
            enabled=False,  # Requires a selected base
            text=N_("&RENAME BASE"),
            anchor=constants.TOP_CENTER,
            autohotkey=True,
            function=self.rename_base,
        )

        self.power_button = button.FunctionButton(
            self,
            (-1, -0.8),
            (-0.3, -0.09),
            autotranslate=True,
            enabled=False,  # Requires a selected base
            text=N_("&POWER STATE"),
            anchor=constants.TOP_RIGHT,
            autohotkey=True,
            function=self.power_state,
        )

        self.new_button = button.FunctionButton(
            self,
            (0, -0.91),
            (-0.3, -0.09),
            autotranslate=True,
            text=N_("&NEW BASE"),
            autohotkey=True,
            function=self.new_base,
        )
        self.destroy_button = button.FunctionButton(
            self,
            (-0.50, -0.91),
            (-0.3, -0.09),
            autotranslate=True,
            enabled=False,  # Requires a selected base
            text=N_("&DESTROY BASE"),
            anchor=constants.TOP_CENTER,
            autohotkey=True,
            function=self.destroy_base,
        )
        self.back_button = button.ExitDialogButton(
            self,
            (-1, -0.91),
            (-0.3, -0.09),
            autotranslate=True,
            text=N_("&BACK"),
            anchor=constants.TOP_RIGHT,
            autohotkey=True,
        )

        self.confirm_destroy = dialog.YesNoDialog(
            self,
            (-0.5, 0),
            (-0.35, -0.7),
            autotranslate=True,
            text=N_("Are you sure you want to destroy this base?"),
            shrink_factor=0.5,
        )

        self.cannot_destroy_last_base = dialog.MessageDialog(
            self,
            autotranslate=True,
            text=N_(
                "Destroying my last active base would be suicidal.  I cannot do that."
            ),
            pos=(-0.5, 0),
            size=(-0.35, -0.7),
            shrink_factor=0.5,
        )

        self.new_base_dialog = NewBaseDialog(self)
        self.location = None

        self.name_dialog = dialog.TextEntryDialog(self)

        self.base_dialog = basescreen.BaseScreen(
            self, (0, 0), anchor=constants.TOP_LEFT
        )

    def make_item(self, canvas):
        canvas.name_display = text.Text(
            canvas,
            (-0.01, -0.05),
            (-0.27, -0.99),
            align=constants.LEFT,
            background_color="clear",
        )
        canvas.base_type = text.Text(
            canvas,
            (-0.27, -0.05),
            (-0.22, -0.99),
            align=constants.LEFT,
            background_color="clear",
        )
        canvas.base_cpu = text.Text(
            canvas,
            (-0.47, -0.05),
            (-0.13, -0.99),
            align=constants.LEFT,
            background_color="clear",
        )
        canvas.status_display = text.Text(
            canvas,
            (-0.59, -0.05),
            (-0.36, -0.99),
            align=constants.LEFT,
            background_color="clear",
        )
        canvas.power_display = text.Text(
            canvas, (-0.90, -0.05), (-0.10, -0.99), background_color="clear"
        )

    def _selection_change(self, *args, **kwargs):
        try:
            current_item = self.listbox.current_item()
        except AttributeError:
            # Occurs when the dialog is created and the list box has not
            # yet been fully initialized.
            return

        if current_item is None:
            base_manipulation_button_state = False
            power_button_state = False
        else:
            base_manipulation_button_state = True
            # "current_item" is just the base name; not the base object itself
            selected_base = self.listbox.key_list[self.listbox.list_pos]
            if len(selected_base.available_power_states) == 1:
                power_button_state = False
            else:
                power_button_state = True
        self.open_button.enabled = base_manipulation_button_state
        self.rename_button.enabled = base_manipulation_button_state
        self.destroy_button.enabled = base_manipulation_button_state
        self.power_button.enabled = power_button_state

    def update_item(self, canvas, name, base):
        if base is None:
            elements = [
                canvas.name_display,
                canvas.base_type,
                canvas.base_cpu,
                canvas.status_display,
                canvas.power_display,
            ]
            for element in elements:
                element.text = ""
        else:
            canvas.name_display.text = name
            canvas.base_type.text = base.spec.name
            canvas.base_cpu.text = ""
            canvas.power_display.text = base.power_state_name
            canvas.power_display.color = state_colors[base.power_state]
            show_cpu = False

            if not base.done:
                canvas.status_display.text = "%s: % 2s%%. %s" % (
                    _("Building Base"),
                    int(base.percent_complete() * 100),
                    _("Completion in %s.") % g.to_time(base.cost_left[2]),
                )
            elif base.spec.force_cpu:
                show_cpu = True
                canvas.status_display.text = ""
            elif base.is_empty():
                canvas.status_display.text = _("Empty")
            elif base.cpus is None:
                canvas.status_display.text = _("Incomplete")
            elif not base.cpus.done:
                canvas.status_display.text = "%s: % 2s%%. %s" % (
                    _("Building CPU"),
                    int(base.cpus.percent_complete() * 100),
                    _("Completion in %s.") % g.to_time(base.cpus.cost_left[2]),
                )
            elif base.is_building_extra():
                show_cpu = True
                canvas.status_display.text = _("Building Item")
            else:
                show_cpu = True
                canvas.status_display.text = _("Complete")

            if show_cpu:
                canvas.base_cpu.text = _("%s CPU") % g.to_money(base.cpu)
            else:
                canvas.base_cpu.text = ""

    def show(self):
        self.listbox.has_focus = True
        return super(LocationScreen, self).show()

    def rebuild(self):
        # Update base location
        if self.location is not None:
            self.location.bases.sort()

            self.listbox.key_list = self.location.bases
            self.listbox.list = [base.name for base in self.location.bases]

            self.name_display.text = self.location.name
            self.modifier_display.text = self.location.get_modifiers_info()

            self.listbox.needs_rebuild = True

        # Ensure that the buttons have the correct state if we have rebuilt
        # the list (or the base state changed in between).
        # Related bug: #282
        self._selection_change()

        # Rebuild dialogs
        self.confirm_destroy.needs_rebuild = True
        self.base_dialog.needs_rebuild = True

        # Update dialog translations
        self.name_dialog.text = _("Enter a name for the base")

        super(LocationScreen, self).rebuild()

    def power_state(self):
        if 0 <= self.listbox.list_pos < len(self.listbox.key_list):
            base = self.listbox.key_list[self.listbox.list_pos]
            base.switch_power()
            self.needs_rebuild = True
            self.parent.needs_rebuild = True

    def destroy_base(self):
        if 0 <= self.listbox.list_pos < len(self.listbox.key_list):
            selected_base = self.listbox.key_list[self.listbox.list_pos]
            all_active_bases = [b for b in g.all_bases() if b.maintains_singularity]
            if len(all_active_bases) == 1 and all_active_bases[0] == selected_base:
                dialog.call_dialog(self.cannot_destroy_last_base, self)
            elif dialog.call_dialog(self.confirm_destroy, self):
                selected_base.destroy()
                self.listbox.list = [b.name for b in self.location.bases]
                self.listbox.key_list = self.location.bases
                self.needs_rebuild = True
                self.parent.needs_rebuild = True

    def open_base(self):
        if 0 <= self.listbox.list_pos < len(self.listbox.key_list):
            base = self.listbox.key_list[self.listbox.list_pos]
            if not base.done:
                return
            self.base_dialog.base = base
            dialog.call_dialog(self.base_dialog, self)
            self.needs_rebuild = True
            self.parent.needs_rebuild = True

    def new_base(self):
        result = dialog.call_dialog(self.new_base_dialog, self)
        if result:
            base_type, base_name = result
            new_base = base.Base(base_type, base_name)
            self.location.add_base(new_base)
            self.needs_rebuild = True
            self.parent.needs_rebuild = True

    def rename_base(self):
        if 0 <= self.listbox.list_pos < len(self.listbox.key_list):
            base = self.listbox.key_list[self.listbox.list_pos]
            self.name_dialog.default_text = base.name
            name = dialog.call_dialog(self.name_dialog, self)
            if name:
                base.name = name
                self.needs_rebuild = True


class NewBaseDialog(dialog.FocusDialog, dialog.ChoiceDescriptionDialog):
    def __init__(
        self,
        parent,
        pos=(0, 0),
        size=(-1, -1),
        anchor=constants.TOP_LEFT,
        *args,
        **kwargs
    ):
        kwargs["yes_type"] = N_("&OK")
        kwargs["no_type"] = N_("&BACK")
        super(NewBaseDialog, self).__init__(parent, pos, size, anchor, *args, **kwargs)
        self.listbox.size = (-0.53, -0.75)
        self.description_pane.size = (-0.45, -0.75)

        self.text_label = text.Text(
            self,
            (0.01, -0.87),
            (-0.25, -0.1),
            autotranslate=True,
            text=N_("Name"),
            anchor=constants.BOTTOM_LEFT,
            borders=(constants.TOP, constants.BOTTOM, constants.LEFT),
            shrink_factor=0.88,
            background_color="pane_background",
        )

        self.text_field = text.EditableText(
            self,
            (-0.26, -0.87),
            (-0.73, -0.1),
            anchor=constants.BOTTOM_LEFT,
            borders=constants.ALL,
            background_color="text_entry_background",
            base_font="normal",
        )

        self.desc_func = self.on_change

        self.yes_button.function = self.finish

    def on_change(self, description_pane, base_type):
        if base_type is not None:
            base_info = base_type.get_info(self.parent.location)
            text.Text(
                description_pane,
                (0, 0),
                (-1, -1),
                text=base_info,
                background_color="pane_background",
                align=constants.LEFT,
                valign=constants.TOP,
                borders=constants.ALL,
            )

            name = generate_base_name(self.parent.location, base_type)
            self.text_field.text = name
            self.text_field.cursor_pos = len(name)

            considered_buyables = []
            fake_base = base.Base("<Undecided>", base_type)
            self.parent.location.modify_base(fake_base)
            considered_buyables.append(fake_base)
            g.pl.considered_buyables = considered_buyables
        else:
            g.pl.considered_buyables = []

    def show(self):
        self.list = []
        self.key_list = []

        base_type_list = sorted(g.base_type.values(), reverse=True)
        for base_type in base_type_list:
            if base_type.available() and base_type.buildable_in(self.parent.location):
                self.list.append(base_type.name)
                self.key_list.append(base_type)

        self._update_desc_pane()
        self.needs_rebuild = True
        self.listbox.has_focus = True
        return super(NewBaseDialog, self).show()

    def handle_update(self, new_item_pos):
        super(NewBaseDialog, self).handle_update(new_item_pos)

    def finish(self):
        if 0 <= self.listbox.list_pos < len(self.key_list):
            type = self.key_list[self.listbox.list_pos]
            name = self.text_field.text
            if not name:
                name = generate_base_name(self.parent.location, type)

            raise constants.ExitDialog((name, type))

    def on_close_dialog(self):
        g.pl.considered_buyables = []


# Generates a name for a base, given a particular location.
def generate_base_name(location, base_type):
    attempts = 0
    name = None
    base_names = {b.name for b in location.bases}
    base_names.add(name)
    while name in base_names:
        # First, decide whether we're going to try significant values or just
        # choose one randomly.
        if random.random() < 0.3:  # 30% chance.
            number = random.choice(g.significant_numbers)
        else:
            number = str(random.randint(0, 32767))

        city = random.choice(location.cities) if location.cities else None
        flavor = random.choice(base_type.flavor) if base_type.flavor else base_type.name

        if city:
            # Translators: Format string for the name of a new base
            # Example: "${NUMBER} ${BASETYPE} in ${CITY}"
            name = _("{CITY} {BASETYPE} {NUMBER}").format(
                CITY=city, BASETYPE=flavor, NUMBER=number
            )
        else:
            # Translators: Name of a new base when location has no cities
            name = _("{BASETYPE} {NUMBER}").format(NUMBER=number, BASETYPE=flavor)

        # Damn translators omitting the ${NUMBER} in template string!
        if attempts > 100:
            name = city + " " + flavor + " " + number
        attempts += 1

    return name
